# -*- encoding: utf-8 -*-
"""hio.help.doming module

Support for Dataclass Object Mapper Dom classes
"""

from __future__ import annotations  # so type hints of classes get resolved later

import functools
import json
import msgpack
import cbor2 as cbor

from typing import Any, Type
from collections.abc import Callable
from dataclasses import dataclass, astuple, asdict, fields, field

# DOM Utilities dataclass utility classes


def dictify(val):
    """
    Returns a serializable dict represention of a dataclass.  If the dataclass
    contains a `_ser` method, use it instead of `asdict`

    Parameters:
         val the dataclass instance to turn into a dict.
    """
    ser = getattr(val, "_ser", None)
    if callable(ser):
        return ser()

    return asdict(val)


def datify(cls, d):
    """Returns instance of dataclass cls converted from dict d. If the dataclass
    cls or any nested dataclasses contains a `_der` method, the use it instead
    of default fieldtypes conversion.

    Parameters:
    cls is dataclass class
    d is dict
    """
    try:
        der = getattr(cls, "_der", None)
        if callable(der):
            return der(d)

        fieldtypes = {f.name: f.type for f in fields(cls)}
        return cls(**{f: datify(fieldtypes[f], d[f]) for f in d})  # recursive
    except Exception:  # Fields in dict d don't match dataclass or something else
        return d  # not a dataclass so end recursion and next level up will process



@dataclass(frozen=True)
class IceMapDom:
    """Base class for frozen dataclasses (codexes) that support map syntax
    Adds support for dunder methods for map syntax dc[name].
    Converts exceptions from attribute syntax to raise map syntax when using
    map syntax.

    Note: iter astuple

    Enables dataclass instances to use Mapping item syntax
    """
    def __getitem__(self, name):
        try:
            return getattr(self, name)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex

    def __iter__(self):
        return iter(self._asdict())

    #def __iter__(self):
        #return iter(astuple(self))


    def _asdict(self):
        """Returns dict version of record"""
        return dictify(self)


    def _astuple(self):
        """Returns dict version of record"""
        return tuple(self._asdict().values())


class MapDom:
    """MapDom is a base class for dataclasses that support map syntax
    Adds support for dunder methods for map syntax dc[name].
    Converts exceptions from attribute syntax to raise map syntax when using
    map syntax.

    Note: iter asdict

    Enables dataclass instances to use Mapping item syntax

    Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d

    Methods:
        __getitem__(self, name)  map like interface
        __setitem__(self, name, value)  map like interface
        __iter__():  asdict(self)
        _asdict(): return self converted to dict
        _astuple(): return self converted to tuple
        _update(*pa,**kwa): update attributes using dict like update syntax
    """

    @classmethod
    def _fromdict(cls, d: dict):
        """returns instance of cls initialized from dict d """
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    def __getitem__(self, name):
        try:
            return getattr(self, name)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex


    def __setitem__(self, name, value):
        try:
            return setattr(self, name, value)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex

    # dataclasses to not allow delattr
    #def __delitem__(self, name):
        #try:
            #return delattr(self, name)
        #except AttributeError as ex:
            #raise IndexError(ex.args) from ex

    def __iter__(self):
        return iter(self._asdict())


    def _asdict(self):
        """Returns dict version of record"""
        return dictify(self)


    def _astuple(self):
        """Returns dict version of record"""
        return tuple(self._asdict().values())


    def _update(self, *pa, **kwa):
        """Use item update syntax
        """
        if len(pa) > 1:
            raise TypeError(f"Expected 1 positional argument got {len(pa)}.")

        if pa:
            di = pa[0]
            if isinstance(di, Mapping):
                for k, v in di.items():
                    self[k] = v
            elif isinstance(di, Iterable):
                for k, v in di:
                    self[k] = v
            else:
                raise TypeError(f"Expected Mapping or Iterable got {type(di)}.")

        for k, v in kwa.items():
            self[k] = v



@dataclass
class RawDom(MapDom):
    """RawDom is subclass of MapDom that adds methods for converting to/from
    raw format serializations in JSON, CBOR, and MGPK. RawDom is a base class
    for such dataclasses. The serialization process is to transform dataclass into
    dict and then serialize into raw. The deserialization process is and to
    deserialize into dict and then convert to dataclass. This allows dataclass
    to stored in databases or to be sent over the wire in messages in raw format.


    Inherited Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d

    Class Methods:
        _fromjson(cls, s: str|bytes): return dataclass converted from json s
        _fromcbor(cls, s: bytes): return dataclass converted from cbor s
        _frommgpk(cls, s: bytes): return dataclass converted from mgpk s

    Inherited Methods:
        __getitem__(self, name)  map like interface
        __setitem__(self, name, value)  map like interface
        __iter__(self): asdict(self)
        _asdict(self): return self converted to dict
        _astuple(): return self converted to tuple
        _update(*pa,**kwa): update attributes using dict like update syntax

    Methods:
        _asjson(self): return bytes self converted to json
        _ascbor(self): return bytes self converted to cbor
        _asmgpk(self): return bytes self converted to mgpk
    """


    @classmethod
    def _fromjson(cls, s: str | bytes):
        """returns instance of clas initialized from json str or bytes s """
        if hasattr(s, "decode"):  # bytes
            s = s.decode() # convert to str
        d = json.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    @classmethod
    def _fromcbor(cls, s: bytes):
        """returns instance of clas initialized from cbor bytes or str s """
        d = cbor.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    @classmethod
    def _frommgpk(cls, s: bytes):
        """returns instance of clas initialized from mgpk bytes or str s """
        d = msgpack.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    def __getitem__(self, name):
        try:
            return getattr(self, name)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex


    def __setitem__(self, name, value):
        try:
            return setattr(self, name, value)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex

    # dataclasses don't allow deletion of attributes
    #def __delitem__(self, name):
        #try:
            #return delattr(self, name)
        #except AttributeError as ex:
            #raise IndexError(ex.args) from ex

    def _update(self, *pa, **kwa):
        """Use item update syntax
        """
        if len(pa) > 1:
            raise TypeError(f"Expected 1 positional argument got {len(pa)}.")

        if pa:
            di = pa[0]
            if isinstance(di, Mapping):
                for k, v in di.items():
                    self[k] = v
            elif isinstance(di, Iterable):
                for k, v in di:
                    self[k] = v
            else:
                raise TypeError(f"Expected Mapping or Iterable got {type(di)}.")

        for k, v in kwa.items():
            self[k] = v


    def __iter__(self):
        return iter(self._asdict())



    def _asdict(self):
        """Returns dict version of record"""
        return dictify(self)


    def _astuple(self):
        """Returns dict version of record"""
        return tuple(self._asdict().values())


    def _asjson(self):
        """Returns json bytes version of record"""
        return json.dumps(self._asdict(),
                          separators=(",", ":"),
                          ensure_ascii=False).encode()



    def _ascbor(self):
        """Returns cbor bytes version of record"""
        return cbor.dumps(self._asdict())



    def _asmgpk(self):
        """Returns mgpk bytes version of record"""
        return msgpack.dumps(self._asdict())



def modify(mods: MapDom|None=None, klas: Type[MapDom]=MapDom)->Callable[..., Any]:
    """Wrapper with argument(s) that injects mods dataclass instance that must
    be a subclass of MapDom. When mods is provided it is injected as the keyword
    arg 'mods' into the wrapped function.
    If mods is None then an instance of klas is created and injected instead.
    This creates a lexical closure of this injection. The value of klas must
    also be a a MapDom subclass.

    Parameters:
        mods (WorkDom|None): default None. Instance of dataclass to be injected
        klas (Type[WorkDom]): defualt WorkDom. Class of dataclass to be injected
                             as default when mods not provided.

    If wrapped function call itself includes mods as an arg whose value is
    not None then does not inject. This allows override of a single call.
    Subsequent calls will resume using the lexical closure or the wrapped
    injected mods whichever was provided.

    Assumes wrapped function defines mods argument as a keyword only parameter
    using '*' such as:
       def f(name='box, over=None, *, mods=None):

    To use inline:
        mods = WorkDom(box=None, over=None, bepre='box', beidx=0)
        g = modify(mods)(f)

    Later calling g as:
        g(name="mid", over="top")
    Actually has mods inject as if called as:
        g(name="mid", over="top", mods=mods)

    Also can be used on a method not just a function.
       def m(self, name='box, over=None, *, mods=None):

    To use inline:
        mods = dict(box=None, over=None, bepre='box', beidx=0)
        m = modify(mods)(self.m)

    Later calling m as:
        m(name="mid", over="top")
    Actually has mods injectd as if called as:
        m(name="mid", over="top", mods=mods)  where self is also injected into method

    Since mods is a mutable collection i.e. dataclass, not an immutable object
    using @decorator syntax on could be problematic as the injected mods would
    be a lexical closure defined in the defining scope not the calling scope.
    Depends on what the use case it for it.

    Example:
        mods = WorkDom(box=None, over=None, bepre='box', beidx=0)
        @modify(mods=mods)
        def f(name="box), over=None, *, mods=None)

    Later calling f as:
        f(name="mid", over="top")
    Actually has mods injected as if called as:
        f(name="mid", over="top", mods=mods)

    But the mods in this case is from the defining scope,not the calling scope.

    Likewise passing in mods=None would result in a lexical closure of mods
    with default values initially that would be shared everywhere f() is called.

    When f() is called with an explicit mods such as f(mods=mods) then that
    call will use the passed in mods not the injected mods. This allows a per
    call override of the injected mods.

    """
    mods = mods if mods is not None else klas()  # lexical closure so not None
    def decorator(f):
        @functools.wraps(f)
        def inner(*pa, **kwa):
            if 'mods' not in kwa or kwa['mods'] is None:  # missing or None
                kwa.update(mods=mods)  # replace or add mods to kwa
            return f(*pa, **kwa)
        return inner
    return decorator




def modize(mods: dict|None=None) -> Callable[..., Any]:
    """Wrapper with argument(s) that injects works dict  when provided as
    keyword arg into wrapped function in order make a boxwork.
    If works is None then creates and injects lexical closure of a dict.
    If wrapped function call itself includes works as an arg whose value is
    not None then does not inject. This allows override of a single call.
    Subsequent calls will resume using the lexical closure or the wrapped
    injected works whichever was provided.

    Assumes wrapped function defines works argument as a keyword only parameter
    using '*' such as:
       def f(name='box, over=None, *, works=None):

    To use inline:
        works = dict(box=None, over=None, bepre='box', beidx=0)
        g = modize(works)(f)

    Later calling g as:
        g(name="mid", over="top")
    Actually has works inject as if called as:
        g(name="mid", over="top", works=works)

    If method then works as wells.
       def f(self, name='box, over=None, *, works=None):

    To use inline:
        works = dict(box=None, over=None, bepre='box', beidx=0)
        f = modize(works)(self.f)

    Later calling f as:
        f(name="mid", over="top")
    Actually has works inject as if called as:
        f(name="mid", over="top", works=works)  the self is automaticall supplied

    Since works is a mutable collection i.e. dict, not an immutable string then
    using it as decorator could be problematic as the works would have lexical
    defining scope not calling scope.
    Which is ok if works has lexical module scope intentionally.

    Example:
        works = dict(box=None, over=None, bepre='box', beidx=0)

        @modize(works=works)
        def f(name="box), over=None, *, works=None)

    Later calling be as:
        f(name="mid", over="top")
    Actually has works inject as if called as:
        f(name="mid", over="top", works=works)

    But the works in this case is from the defining scope of be not the calling
    scope.

    Likewise passing in works=None would result in a lexical closure of works
    with default values initially that would be shared everywhere f() is called
    with whatever values each call changes in the lexical closure of works.
    """
    mods = mods if mods is not None else {}  # {} in lexical closure when None
    def decorator(f):
        @functools.wraps(f)
        def inner(*pa, **kwa):
            if 'mods' not in kwa or kwa['mods'] is None:  # missing or None
                kwa.update(mods=mods)  # replace or add works to kwa
            return f(*pa, **kwa)
        return inner
    return decorator




# -*- coding: utf-8 -*-
"""
hio.hioing module

Generic Constants and Classes
Exception Classes

"""
import sys
from collections import namedtuple

Versionage = namedtuple("Versionage", "major minor")

Version = Versionage(major=1, minor=0)  # Protocol Version

SEPARATOR =  "\r\n\r\n"
SEPARATOR_BYTES = SEPARATOR.encode("utf-8")


class Mixin():
    """
    Base class to enable consistent MRO for mixin multiple inheritance
    Allows each subclass to call
    super(MixinSubClass, self).__init__(*pa, **kwa)
    So the __init__ propagates to common top of Tree
    https://medium.com/geekculture/cooperative-multiple-inheritance-in-python-practice-60e3ac5f91cc
    """
    def __init__(self, *pa, **kwa):
        pass


class HioError(Exception):
    """
    Base Class for hio exceptions

    To use   raise HioError("Error: message")
    """

class SizeError(HioError):
    """
    Resource size related errors
    Usage:
        raise SizeError("error message")
    """

class ValidationError(HioError):
    """
    Validation related errors
    Usage:
        raise ValidationError("error message")
    """


class VersionError(ValidationError):
    """
    Bad or Unsupported Version

    Usage:
        raise VersionError("error message")
    """

class OglerError(HioError):
    """
    Error using or configuring Ogler

    Usage:
        raise OglerError("error message")
    """

class FilerError(HioError):
    """
    Error using or configuring Filer

    Usage:
        raise FilerError("error message")
    """

class NamerError(HioError):
    """
    Error using or configuring Namer

    Usage:
        raise NamerError("error message")
    """

class MemoerError(HioError):
    """
    Error using or configuring Memoer

    Usage:
        raise MemoGramError("error message")
    """

class MultiError(HioError):
    """
    Error using or configuring multiprocessing support classes

    Usage:
        raise MultiError("error message")
    """

class HierError(HioError):
    """
    Error using or configuring hiering support classes

    Usage:
        raise HierError("error message")
    """



# -*- encoding: utf-8 -*-
"""
hio.core.multidoing Module

Provides multiprocessing support



"""
import os
import sys
import time
import logging
import json
import signal
import re
import multiprocessing as mp

from collections import deque, namedtuple
from dataclasses import dataclass, astuple, asdict, field
import typing


from . import tyming
from .doing import Doist, Doer
from .. import hioing
from .. import help
from ..help import ogling, timing, helping, Namer, RawDom, IceMapDom

from ..core.uxd import PeerMemoer

ogler = ogling.initOgler(prefix='hio_mp', name="Boss", level=logging.ERROR)





"""Bossage (namedtuple):
Bosser info to be injected in Crewer as Bosser enter time
Fields:
   name (str | None): name of Boss for resource management
   path (str | None): UXD cmd memo path used by crew to talk to their boss
"""
Bossage = namedtuple("Bossage", "name path")

"""Loadage (namedtuple):
Bosser info to be injected into Crewer .start() containing both crew doist
parms for Process target kwargs and and Crewer parms

Fields:
   name (str): child doist identifier for resources.
   tyme (float): child doist start tyme
   tock (float | None): child doist tock, tyme lag between runs
   real (bool): child doist True means run in real time, tyme is real time
   limit (float | None): child doist tyme limit. None means run forever
   doers (list[Doers]): child doist List of Doers
   temp (bool | None): True means use temporary file resources
   boss (Bossage | None): Bosser  info for Crewer. May be filled at enter time
                      Crewer uses to contact Bosser.
"""
Loadage = namedtuple("Loadage", "name tyme tock real limit doers temp boss")


# Regular expression to extract tag from JSON serializations of Bosser Crewer memos
TAGREX = r'^\{"tag":"(?P<tag>[A-Z]*)"'
# Usage: if retag.match(memo): or if not Reb64.match(memo): memo is str
# tag = retag.match(memo).group(1)
# tag = retag.match(memo).group("tag")
# if match := retag.match(memo): tag =match.group(0)
Retag = re.compile(TAGREX)  # compile is faster

"""RawDom hidden methods

    Inherited Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d
        _fromjson(cls, s: str|bytes): return dataclass converted from json s
        _fromcbor(cls, s: bytes): return dataclass converted from cbor s
        _frommgpk(cls, s: bytes): return dataclass converted from mgpk s

    Inherited Methods:
        __iter__(self): asdict(self)
        _asdict(self): return self converted to dict
        _asjson(self): return bytes self converted to json
        _ascbor(self): return bytes self converted to cbor
        _asmgpk(self): return bytes self converted to mgpk

"""
@dataclass
class HandDom(RawDom):
    """Configuration dataclass of Crewer crew hand info managed by its Bosser
    boss.

    Attributes:
        proc (typing.Any | None): crew hand subprocess or None
        exiting (bool): True means commanded to exit but may not have exited yet
                        False means not yet commanded to exit

    """
    proc: typing.Any = None  # crew hand subprocess
    exiting: bool = False  # True means commanded to exit but may not have exited yet


@dataclass
class CrewDom(RawDom):
    """Configuration dataclass of crew subprocess Doist parameters for crew doist
    and its Crewer to be injected by Bosser.
    Use this when storing configuration in database or file. Use RawDom
    serialization hidden methods:


    Attributes:
        name (str): child doist identifier for resources.
        tyme (float): child doist start tyme
        tock (float | None): child doist tock, tyme lag between runs
        real (bool): child doist True means run in real time, tyme is real time
        limit (float | None): child doist tyme limit. None means run forever
        doers (list[Doers]): child doist List of Doers
        temp (bool | None): True means use temporary file resources
        boss (Bossage | None): Bosser  info for Crewer. May be filled at enter time
                                Crewer uses to contact Bosser.
    """
    name: str ='child'  # unique identifier of child process and associated resources
    tyme: float = 0.0    # child doist start tyme
    tock: float | None = None  # child doist tyme lag between runs, None means ASAP
    real: bool = True  # child doist True means run in real time, tyme is real time
    limit: float | None = None  # child doist tyme limit. None mean run forever
    doers: list = field(default_factory=list)  # child doist list of doers
    temp: bool | None = False  # use temporary file resources if any
    boss: (Bossage | None) = Bossage(name=None, path=None)  # Bosser info


@dataclass
class MemoDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for control memos
    Between Boss and Crew Doers via their .peer UXD BossMemoer or CrewMemoer.

    Attributes:
        tag (str): type of memo
        name (str): unique identifier as source of memo
        load (dict): type specific payload of memo
    """
    tag: str = 'REG'    # type of memo
    name: str ='hand'  # unique identifier of source
    load: dict = field(default_factory=dict)  # type specific payload


@dataclass
class RegDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for REG memos
    Between Boss and Crew Doers via their .peer UXD BossMemoer or CrewMemoer.

    Attributes:
        tag (str): type of memo
        name (str): unique identifier as source of memo
        load (dict): empty dict
    """
    tag: str = 'REG'    # type of memo
    name: str ='hand'  # unique identifier of source
    load: dict = field(default_factory=dict)  # empty dict


@dataclass
class AddrDom(RawDom):
    """Load Field Value of ACK

    Attributes:
        tag (str): type of memo being acked
        name (str): unique identifier as source of memo being acked
        addr (dict): addr of source of memo being acked
    """
    tag: str = 'REG'    # type of memo being acked
    name: str ='hand'  # unique identifier of source of ack
    addr: str = ''  # addr of source of acked memo


@dataclass
class AckDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for ACK memos
    Between Boss and Crew Doers via their .peer UXD BossMemoer or CrewMemoer.
    The load of the ACK is an AddrDom instance with the fields, tag, name, and addr.
    The AddrDom tag and name fields come from the memo being acked. The addr
    field is the src address path of the memo being acked.

    Attributes:
        tag (str): type of memo
        name (str): unique identifier as source of memo
        load (AddrDom): attribution info of acked memo
    """
    tag: str = 'ACK'    # type of memo
    name: str ='boss'  # unique identifier of source
    load: AddrDom = field(default_factory=AddrDom)  # instance of AddrDom


@dataclass
class EndDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for END memos
    Sent by Boss to its Crew Doers to gracefully terminate. Sent via their .peer
    UXD BossMemoer or CrewMemoer instead of using an OS SIGTERM signal.

    Attributes:
        tag (str): type of memo
        name (str): unique identifier of boss
        load (dict): empty dict
    """
    tag: str = 'END'    # type of memo
    name: str ='boss'  # unique identifier of boss
    load: dict = field(default_factory=dict)  # empty dict


@dataclass
class BokDom(RawDom):
    """Inter Boss Crew Hand structured memo dataclass. Used for BOK memos
    Sent by Boss to its Crew Doers with an address book of the crew hands. This
    enables crew hand to crew hand peer-to-peer messages sent via their .peer
    UXD CrewMemoer.

    The load value is a dict keyed by crew hand names with values that are the
    crew hand UXD peer addres file paths. The items of the dict are (name, addr)
    tuples

    Attributes:
        tag (str): type of memo
        name (str): unique identifier of boss
        load (dict): items are (name, addr) tuples
    """
    tag: str = 'BOK'    # type of memo
    name: str ='boss'  # unique identifier of boss
    load: dict = field(default_factory=dict)  # needs to be filled


@dataclass(frozen=True)
class TagDomCodex(IceMapDom):
    """Codex keyed by memo tag with value of associated MemoDom subclass.
    Attribute values are classes not instances

    iters asdict

    Attributes:
        REG (type[RegDom]): RegDom
        ACK (type[AckDom]): AckDom
        END (type[EndDom]): EndDom
        BOK (type[BokDom]): BokDom

    """
    REG: type[RegDom] = RegDom  # value is class not instance
    ACK: type[AckDom] = AckDom  # value is class not instance
    END: type[EndDom] = EndDom  # value is class not instance
    BOK: type[BokDom] = BokDom  # value is class not instance

TagDex = TagDomCodex()  # make instance



class MultiDoerBase(Namer, PeerMemoer, Doer):
    """MultiDoerBase is base class for Doers in multiprocessing. Each subclass
    has support for UXD peer communications via PeerMemoer super class. As well
    as support for local logging given scope of module global ogler at enter time.

    See Namer, PeerMemoer, and Doer for inherited attributes, properties, and methods.

    Inherited Class Attributes:
        See Namer, PeerMemoer and Doer Classes

    Class Attributes:
        Tagex (TagDomCodex):  codex mapping memo tags to memo doms

    Inherited Attributes:  (See Doer and PeerMemoer for all)
        done (bool): completion state:
                     True means completed fully. Otherwise incomplete.
                     Incompletion value may be None or False.
        opts (dict): schedulaer injects options from .opts into its .do generator
                     function as **opts parameter.
        name (str): unique identifier for this instance used to manage local resources
        temp (bool): True means logger or other file resources created by
                        .start() will use temp
                     False otherwise
        reopen (bool): True (re)open with this init
                       False not (re)open with this init but later (default)
        bc (int | None): count of transport buffers of MaxGramSize

    Attributes:
        logger (Logger | None): from module scope ogler created at enter time
                        with local resources.
        graceful (bool): indication to gracefully exit on next recur. Set by
                         .force signal handler. Must add check in .self.recur
                         if self.graceful: sys.exit()
                         True means exit on next recur
                         False otherwise.

    Inherited Properties:
        tyme (float): is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (closure): function wrapper closure returned by Tymist.tymen()
                        method. When .tymth is called it returns associated
                        Tymist.tyme. Provides injected dependency on Tymist
                        tyme base.
        tock (float): desired time in seconds between runs or until next run,
                 non negative, zero means run asap
        addrByName (dict): mapping between (name, address) pairs, these
            must be one-to-one so that inverse is also one-to-one
        nameByAddr (dict): mapping between (address, name) pairs, these
            must be one-to-one so that inverse is also one-to-one
        countNameAddr (int): count of entries in .addrByName


    Properties:

    Inherited Methods:
        __call__()  makes instance callable as generator function returning generator
        do() generator method that returns generator
        enter() is enter context action method
        recur() recur context action method or generator method
        clean() clean context action method
        exit() exit context method
        close() close context method
        abort() abort context method
        wind()  injects ._tymth dependency from associated Tymist to get its .tyme
        clearAllNameAddr()
        getAddr(name)
        getName(addr)
        addNameAddr(name, addr)
        remNameAddr(name=None, addr=None)
        changeAddrAtName(name=None, addr=None)
        changeNameAtAddr(addr=None, name=None)

    """
    Tagex = TagDex

    def __init__(self, *, name='base', temp=False, reopen=False, bc=4, **kwa):
        """Initialize instance.

        Inherited Parameters:  (see Doer and PeerMemoer for all)
            name (str): unique identifier for this Bosser boss to be used
                        to manage local resources
            temp (bool): True means logger or other file resources created by
                            .start() will use temp
                         False otherwise
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            bc (int | None): count of transport buffers of MaxGramSize

        Parameters:


        """
        super(MultiDoerBase, self).__init__(name=name,
                                            temp=temp,
                                            reopen=reopen,
                                            bc=bc,
                                            **kwa)

        self.logger = None  # assign later from ogler in enter time/scope
        self.graceful = False  # indiction by .force signal handler to exit on next recur



    def force(self, signum, frame):  # signal handler for forced but graceful exit
        self.graceful = True  # exit gracefully in recur
        self.logger.debug("Caught signal=%d so graceful forced exit.", signum)


    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Start processes with config from .loads
        Not a generator method.

        Parameters:
            temp (bool | None): True means use temporary file resources if any.
                                None means ignore parameter value. Use self.temp.

        Inject temp or self.temp into file resources here if any
        Doist or DoDoer winds its doers on enter

        """
        self.logger = ogler.getLogger()  # uses ogler in enter scope
        signal.signal(signal.SIGINT, self.force)  # register signal handers
        signal.signal(signal.SIGTERM, self.force) # register signal handers


    def recur(self, tyme):
        """Do 'recur' context.
        Must be overidden in subclass
        """
        if self.graceful:  # signal handler.force caught signal so exit here
            sys.exit()

        self.service()

        return False  # this  is just a default

    @staticmethod
    def tojson(d):
        """Returns compact JSON serialization of d suitable for .memoit.

        Parameters:
            d (dict | list): object to be serialized

        Returns:
            s (str): serialized JSON
        """
        return json.dumps(d, separators=(",", ":"),ensure_ascii=False )


class Bosser(MultiDoerBase):
    """Bosser spawns multiple crew hand subprocesses and injects each with
    a Doist and Doers. The boss Doists runs the Bosser in the parent process.
    Each crew hand Doist runs a Crewer that coordinates with the Bosser.

    Analogy Boss runs a Crew of Hans. The parent process has a boss Doist which
    runs a Bosser. Each crew hand is a child process with its own crew doist
    that runs its own Crewer

    See MultiDoerBase for all inherited attributes, properties, and methods.

    Inherited Class Attributes:
        See MultiDoerBase Class


    Inherited Attributes:  (See MultiDoerBase Class for all)
        name (str): unique identifier for this instance
                    used to manage local resources
        logger (Logger | None): from module scope ogler created at enter time
                        with local resources.
        graceful (bool): indication to gracefully exit on next recur. Set by
                         .force signal handler. Must add check in .self.recur
                         if self.graceful: sys.exit()
                         True means exit on next recur
                         False otherwise.

    Attributes:
        loads (list[dict]): Bosser info to be injected into Crewer .start()
                            containing both crew doist parmss for Process target
                            kwargs and and Crewer parms
                            (see Loadage._asdict() or CrewDom._asdict())
        ctx (mp.context.SpawnContext | None): context under which to spawn processes
        crew (dict): values HandDom instances keyed by name
        crewed (bool): True means all crew members have registered memo interface
                            with this boss.
                       False means not yet


    Inherited Properties:
        See MultiDoerBase Class

    Properties:

    Inherited Methods:
        See MultiDoerBase Class
    """

    def __init__(self, *, name='boss',loads=None, **kwa):
        """Initialize instance.

        Inherited Parameters:  (see Doer and PeerMemoer for all)
            name (str): unique identifier for this instance to be used
                        to manage local resources

        Parameters:
            loads (list[dict]): parameters used to spinup crew hand subprocess
                                .start(). See fields of Loadage and Bossage

        """
        super(Bosser, self).__init__(name=name, **kwa)
        self.loads = loads if loads is not None else []
        self.ctx = mp.get_context('spawn')
        self.crew = {}  # dict of HandDom instances keyed by crew name
        self.crewed = False  # True means crew successfully registered with boss



    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Start processes with config from .loads
        Not a generator method.

        Parameters:
            temp (bool | None): True means use temporary file resources if any.
                                None means ignore parameter value. Use self.temp.

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        super(Bosser, self).enter(temp=temp)

        self.logger.debug("Bosser Enter: name=%s size=%d, ppid=%d, pid=%d, "
                          "module=%s, temp=%s, ogler=%s tyme=%f.",
                          self.name, len(self.loads), os.getppid(), os.getpid(),
                          __name__, temp, ogler.name, self.tyme)

        self.reopen(temp=temp)
        self.logger.debug("Boss name=%s path=%s opened=%s.",
                    self.name, self.path, self.opened)
        boss = Bossage(name=self.name, path=self.path)  # use peer at enter

        for load in self.loads:
            name = load["name"]
            doers = load["doers"]
            if not doers or not isinstance(doers[0], Crewer):  # first doer must be Crewer
                raise hioing.MultiError(f"No crew doer for crew hand doist={name}.")

            doers[0].boss = boss  # assign boss contact info to crew doer
            doers[0].name = name  # make name of Crewer same as name of crew doist

            if name not in self.crew:  # ensure unique by name
                proc = self.ctx.Process(name=name,
                                         target=self.start,
                                         kwargs=load)
                dom = HandDom(proc=proc, exiting=False)
                self.crew[proc.name] = dom
                proc.start()
            else:
                raise hioing.MultiError(f"Non-unique crew hand {name=} in loads.")



    def recur(self, tyme):
        """Do 'recur' context."""
        #self.logger.debug("Bosser Recur: name=%s, pid=%d, ogler=%s, tyme=%f.",
                          #self.name, os.getpid(), ogler.name, tyme)

        if self.graceful:  # signal handler.force caught signal so exit here
            sys.exit()

        self.service()

        if self.crewed:
            if not self.ctx.active_children():
                return True  # childred exited on their own so unforced exit

        # otherwise run forever

        return False  # incomplete recur again


    def exit(self):
        """Do 'exit' (try finally) context."""
        self.logger.debug("Bosser Exit: name=%s, ppid=%d, pid=%d, module=%s, "
                          "ogler=%s, tyme=%f.", self.name, os.getppid(),
                          os.getpid(), __name__, ogler.name, self.tyme)

        self.close(clear=True)

        self.logger.debug("Boss name=%s path=%s opened=%s.",
                self.name, self.path, self.opened)

        for name, dom in self.crew.items():  # dom is CrewDom instance
            if dom.proc.is_alive():
                dom.proc.terminate()  # force exit with SIGTERM
                self.logger.debug("Boss name=%s pid=%d terminating hand name=%s "
                                  "pid=%d.", self.name, os.getpid(), name,
                                  dom.proc.pid)


    def cease(self):
        """Do 'cease' context."""


    def abort(self, ex):
        """Do 'abort' context."""


    def serviceRxMemos(self):
        """Service all memos in .rxms (greedy) if any

        Override in subclass to handle result(s) and put them somewhere
        """
        while self.rxms:
            memo, src, vid = self._serviceOneRxMemo()
            self.logger.debug("Boss Peer RX: name=%s rx from src=%s memo=%s.",
                                self.name, src, memo)

            if not (match := Retag.match(memo)):  # missing or unreconizable tag
                continue  # drop memo

            tag = match.group("tag")
            if tag not in self.Tagex:  # unrecognized tag
                continue  # so drop memo

            try:
                mdom = self.Tagex[tag]._fromjson(memo)
            except ValueError:  # invalid formatted json for dom at tag
                continue  # so drop memo

            if tag == "REG":
                name = mdom.name
                try:  # add to address book
                    self.addNameAddr(name=name, addr=src)
                except hioing.NamerError as ex:
                    self.changeAddrAtName(name=name, addr=src)
                # Send ACK to REG to src crew hand
                dst = src
                mack = AckDom(name=self.name, load=AddrDom(name=name, addr=src))._asjson().decode()
                self.memoit(mack, dst)

                if self.countNameAddr == len(self.crew):
                    self.crewed = True
                    self.logger.debug("Boss name=%s crewed=%s with size=%d at "
                                      "tyme=%f.", self.name, self.crewed,
                                      len(self.crew), self.tyme)

                    # send address book of crew hands to all crew hands
                    book = self.addrByName  # dict of (name, addr) pairs
                    mbok = BokDom(name=self.name, load=book)._asjson().decode()
                    for name, dom in self.crew.items():  # dom is CrewDom instance
                        if dom.proc.is_alive():
                            dst = self.getAddr(name=name)
                            self.memoit(mbok, dst)



    @staticmethod
    def start(*, name='crew', tyme=0.0, tock=None, real=True, limit=None,
               doers=None, temp=None, boss=None):
        """Process target function to spinup and run doist inside crew subprocess
        after it has been started.

        Parameters:
            name (str): unique crew hand name to be used to manage resources
            tyme (float): crew doist initial value of cycle time in seconds
            tock (float | None): crew doist tock time in seconds. None means run Asap
            real (bool): crew doist True means run in real time,
                        Otherwise run faster than real
            limit (float | None): crew doist seconds for max run time of doist.
                                  None means no limit.
            doers (iterable[Doer] | None): crew doist Doer class instances
                                   First entry must be Crewer
            temp (bool | None): True means use temp file resources by injection.
                                Otherwise ignore do not inject.
            boss (Bossage | None): boss info. May be filled at enter time
                                  Crewer uses to contact Bosser.


        Doist must be built after process started so local tymth closure is created
        inside subprocess so that when doist winds the tyme for its doers
        the tymth closure is locally sourced.

        When run inside subprocess, this static method provides the outside
        scope for any Doist and doers that reference objects inside the
        __main__ moduleo that calls this method such as ogler.
        The subprocess gets a picked copy of these module globals. Any Doer
        inside then has access to the module globals and can update them as
        neede. In the case of ogler this means changing ogler.level, ogler.temp
        and running ogler.reopen(temp=temp) as appropriate.
        """
        logger = ogler.getLogger()  # uses ogler from subprocess scope

        logger.debug("Crew Start: name=%s, ppid=%d, pid=%s, module=%s, temp=%s, ogler=%s.",
                        name, os.getppid(), os.getpid(), __name__, temp, ogler.name)
        time.sleep(0.01)

        doist = Doist(name=name, tyme=tyme, tock=tock, real=real, limit=limit,
                      doers=doers, temp=temp)
        try:
            doist.do()

        finally:
            logger.debug("Crew End: name=%s, ppid=%d, pid=%s, module=%s, temp=%s, "
                         "ogler=%s, doist.tyme=%f.", name, os.getppid(), os.getpid(),
                         __name__, temp, ogler.name, doist.tyme)



class Crewer(MultiDoerBase):
    """Crewer runs interface between a given crew hand subprocess and its
    boss process. This must be first doer run by crew hand subprocess doist.

    See MultiDoerBase for all inherited attributes, properties, and methods.

    Inherited Class Attributes:
        See MultiDoerBase Class


    Inherited Attributes:  (See MultiDoerBase Class for all)
        name (str): unique identifier for this instance
                    used to manage local resources
        logger (Logger | None): from module scope ogler created at enter time
                        with local resources.
        graceful (bool): indication to gracefully exit on next recur. Set by
                         .force signal handler. Must add check in .self.recur
                         if self.graceful: sys.exit()
                         True means exit on next recur
                         False otherwise.

    Attributes:
        boss (Bossage | None): contact info for communicating with boss
        registered (bool): True means .path acked registered with boss memoing
                           False not yet registered


    Inherited Properties:
        See MultiDoerBase Class

    Properties:

    Inherited Methods:
        See MultiDoerBase Class

    """

    def __init__(self, *, name='crew', boss=None, **kwa):
        """Initialize instance.

        Inherited Parameters:
            name (str): unique identifier for this instance to be used
                        to manage local resources
            temp (bool): True means logger or other file resources created by

        Parameters:
            boss (Bossage): contact info for Bosser. assigned by boss at enter


        """
        super(Crewer, self).__init__(name=name, **kwa)
        self.boss = boss if boss is not None else Bossage(name=None, path=None)
        self.registered = False  # True means .path acked registered with boss memoing


    def force(self, signum, frame):  # signal handler for forced but graceful exit
        self.graceful = True  # exit gracefully in recur
        self.logger.debug("Doer name=%s caught signal=%d so graceful forced exit.",
                          self.name, signum)


    def enter(self, *, temp=None):
        """Do 'enter' context.
        Set up resources. Comparable to context manager enter.
        Not a generator method.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        super(Crewer, self).enter(temp=temp)

        self.logger.debug("Crewer Enter: hand name=%s pid=%d, temp=%s, ogler=%s, "
                          "tyme=%f.", self.name, os.getpid(), temp, ogler.name,
                          self.tyme)

        self.logger.debug("Hand name=%s Boss: name=%s path=%s.",
                            self.name, self.boss.name, self.boss.path)
        self.addNameAddr(name=self.boss.name, addr=self.boss.path)

        self.reopen(temp=temp)

        self.logger.debug("Hand name=%s path=%s opened=%s.",
                    self.name, self.path, self.opened)

        memo = RegDom(name=self.name)._asjson().decode()
        dst = self.boss.path
        self.memoit(memo, dst)



    def recur(self, tyme):
        """Do 'recur' context."""
        self.logger.debug("Crewer Recur: hand name=%s, pid=%d, ogler=%s, tyme=%f.",
                          self.name, os.getpid(), ogler.name, tyme)
        if self.graceful:  # signal handler.force caught signal so exit here
            sys.exit()

        self.service()


        return False  # incomplete


    def exit(self):
        """Do 'exit' (try finally) context."""
        self.close(clear=True)

        self.logger.debug("Crewer Exit: hand name=%s pid=%d, ogler=%s, tyme=%f.",
                          self.name, os.getpid(), ogler.name, self.tyme)
        self.logger.debug("Hand name=%s path=%s opened=%s.",
                    self.name, self.path, self.opened)


    def cease(self):
        """Do 'cease' context."""



    def abort(self, ex):
        """Do 'abort' context."""



    def serviceRxMemos(self):
        """Service all memos in .rxms (greedy) if any

        Override in subclass to handle result(s) and put them somewhere
        """
        while self.rxms:
            memo, src, vid = self._serviceOneRxMemo()
            self.logger.debug("Hand Peer RX: name=%s rx from src=%s memo=%s.",
                                self.name, src, memo)

            if not (match := Retag.match(memo)):  # missing or unreconizable tag
                continue  # drop memo

            tag = match.group("tag")
            if tag not in self.Tagex:  # unrecognized tag
                continue  # so drop memo

            try:
                mdom = self.Tagex[tag]._fromjson(memo)
            except ValueError:  # invalid formatted json for dom at tag
                continue  # so drop memo

            if tag == "END":
                name = mdom.name
                if name == self.boss.name and src == self.boss.path:
                    self.logger.debug("Hand name=%s exiting.", self.name)
                    self.graceful = True  # next recur graceful exit

            elif tag == "ACK":
                name = mdom.name
                if name == self.boss.name and mdom.load.tag == 'REG':
                    self.registered = True
                    self.logger.debug("Hand name=%s registered with boss=%s",
                                        self.name, self.boss.name)

            elif tag == "BOK":
                name = mdom.name
                if name == self.boss.name:
                    load = mdom.load
                    self.logger.debug("Hand name=%s got from boss=%s "
                                      "address book update of other crew"
                                      " hands if any.",
                                      self.name, self.boss.name)
                    for name, addr in load.items():
                        if name != self.name:  # don't put self in own address book
                            self.addNameAddr(name=name, addr=addr)






# -*- encoding: utf-8 -*-
"""
hio.core.udp.udping Module
"""
import sys
import platform
import os
import errno
import socket
from contextlib import contextmanager

from ... import hioing
from .. import coring
from ...base import tyming, doing
from ... import help

logger = help.ogler.getLogger()


UDP_MAX_DATAGRAM_SIZE = (2 ** 16) - 1  # 65535
UDP_MAX_SAFE_PAYLOAD = 548  # IPV4 MTU 576 - udp headers 28
# IPV6 MTU is 1280 but headers are bigger
UDP_MAX_PACKET_SIZE = min(1024, UDP_MAX_DATAGRAM_SIZE)  # assumes IPV6 capable equipment



class Peer(hioing.Mixin):
    """Class to manage non blocking I/O on UDP socket.

    Attributes:
        name (str): unique identifier of peer for managment purposes
        ha (tuple): host address of form (host,port) of type (str, int) of this
                peer's socket address.
        bs (int): buffer size
        wl (WireLog): instance ref for debug logging of over the wire tx and rx
        bcast (bool): True enables sending to broadcast addresses from local socket
                      False otherwise
        ls (socket.socket): local socket of this Peer
        opened (bool): True local socket is created and opened. False otherwise

    Properties:
        host (str): element of .ha duple
        port (int): element of .ha duple


    """

    def __init__(self, *,
                 name='main',
                 ha=None,
                 host='',
                 port=55000,
                 bufsize=1024,
                 wl=None,
                 bcast=False,
                 **kwa):
        """
        Initialization method for instance.
        Parameters:
            name (str): unique identifier of peer for managment purposes
            ha (tuple): local socket (host, port) address duple of type (str, int)
            host (str): address where '' means any interface on host
            port (int): socket port
            bs (int):  buffer size
            wl (WireLog): instance to log over the wire tx and rx
            bcast (bool): True enables sending to broadcast addresses from local socket
                          False otherwise
        """
        super(Peer, self).__init__(**kwa)
        self.name = name
        self.ha = ha or (host, port)  # ha = host address duple (host, port)
        host, port = self.ha
        host = coring.normalizeHost(host)  # ip host address
        self.ha = (host, port)

        self.bs = bufsize
        self.wl = wl
        self.bcast = bcast

        self.ls = None  # local socket for this Peer
        self.opened = False


    @property
    def host(self):
        """
        Property that returns host in .ha duple
        """
        return self.ha[0]


    @host.setter
    def host(self, value):
        """
        setter for host property
        """
        self.ha = (value, self.port)


    @property
    def port(self):
        """
        Property that returns port in .ha duple
        """
        return self.ha[1]


    @port.setter
    def port(self, value):
        """
        setter for port property
        """
        self.ha = (self.host, value)


    def actualBufSizes(self):
        """Returns duple of the the actual socket send and receive buffer size
        (send, receive)
        """
        if not self.ls:
            return (0, 0)

        return (self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF),
                self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))

    def open(self):
        """Opens socket in non blocking mode.

        if socket not closed properly, binding socket gets error
           OSError: (48, 'Address already in use')
        """
        #create socket ss = server socket
        self.ls = socket.socket(socket.AF_INET,
                                socket.SOCK_DGRAM,
                                socket.IPPROTO_UDP)

        if self.bcast:  # needed to send broadcast, not needed to receive
            self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        # make socket address and port reusable. doesn't seem to have an effect.
        # the SO_REUSEADDR flag tells the kernel to reuse a local socket in
        # TIME_WAIT state, without waiting for its natural timeout to expire.
        # Also use SO_REUSEPORT on linux and darwin
        # https://stackoverflow.com/questions/14388706/how-do-so-reuseaddr-and-so-reuseport-differ

        self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if platform.system() != 'Windows':
            self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

        # setup buffers
        if self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF) <  self.bs:
            self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, self.bs)
        if self.ls.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF) < self.bs:
            self.ls.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, self.bs)
        self.ls.setblocking(0)  # non blocking socket

        #bind to Host Address Port
        try:
            self.ls.bind(self.ha)
        except OSError as ex:
            logger.error("Error opening UDP %s\n %s\n", self.ha, ex)
            return False

        self.ha = self.ls.getsockname()  # get resolved ha after bind
        self.opened = True
        return True

    def reopen(self, **kwa):
        """Idempotently open socket
        """
        self.close()
        return self.open()

    def close(self):
        """Closes  socket.
        """
        if self.ls:
            self.ls.close() #close socket
            self.ls = None
            self.opened = False

        return not self.opened  # True means closed successfully

    def receive(self, **kwa):
        """Perform non blocking read on  socket.

        Returns:
            tuple of form (data, sa)
            if no data then returns (b'',None)
            but always returns a tuple with two elements
        """
        try:
            data, sa = self.ls.recvfrom(self.bs)  # sa is source (host, port)
        except OSError as ex:
            # ex.args[0] == ex.errno for better compat
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if (ex.args[0] in (errno.EAGAIN,
                              errno.EWOULDBLOCK)):
                return (b'', None) #receive has nothing empty string for data
            else:
                logger.error("Error receive on UDP %s\n %s\n", self.ha, ex)
                raise #re raise exception ex

        if self.wl:  # log over the wire receive
            self.wl.writeRx(data, who=sa)

        return (data, sa)

    def send(self, data, dst, **kwa):
        """Perform non blocking send on  socket.

        Returns:
            cnt (int): count of bytes actually sent, may be less than len(data).

        Parameters:
            data (bytes):  payload to send
            dst (str): udp destination addr duple of form (host: str, port: int)
        """
        try:
            cnt = self.ls.sendto(data, dst)  # count == int number of bytes sent
        except OSError as ex:
            # ex.args[0] == ex.errno for better compat
            # the value of a given errno.XXXXX may be different on each os
            # EAGAIN: BSD 35, Linux 11, Windows 11
            # EWOULDBLOCK: BSD 35 Linux 11 Windows 140
            if (ex.args[0] in (errno.EAGAIN,
                               errno.EWOULDBLOCK,
                               errno.ENOBUFS,
                               errno.ENOMEM)):
                # not enough buffer space to send, do not consume data
                return 0  # try again later with same data

            else:
                logger.error("Error send UDP from %s to %s.\n %s\n", self.ha, dst, ex)
                cnt = 0
            raise

        if self.wl:  # log over the wire actually sent data portion
            self.wl.writeTx(data[:cnt], who=dst)

        return cnt




@contextmanager
def openPeer(cls=None, name="test", **kwa):
    """
    Wrapper to create and open UDP Peer instances
    When used in with statement block, calls .close() on exit of with block

    Parameters:
        cls (Class): instance of subclass instance
        name (str): unique identifer of peer. Enables management of Peer sockets
                    by name.
    Usage:
        with openPeer() as peer0:
            peer0.receive()

        with openPeer(cls=PeerBig) as peer0:
            peer0.receive()

    """
    peer = None

    if cls is None:
        cls = Peer
    try:
        peer = cls(name=name, **kwa)
        peer.reopen()

        yield peer

    finally:
        if peer:
            peer.close()



class PeerDoer(doing.Doer):
    """Basic UDP Peer Doer
    Stub Override in Subclass
    To test in WingIde must configure Debug I/O to use external console
    See Doer for inherited attributes, properties, and methods.

    See Doer for inherited attributes, properties, and methods.

    Attributes:
       .peer is UDP Peer instance

    """

    def __init__(self, peer, **kwa):
        """
        Initialize instance.

        Parameters:
           peer is UXD Peer instance
        """
        super(PeerDoer, self).__init__(**kwa)
        self.peer = peer


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        self.peer.reopen(temp=temp)


    def recur(self, tyme):
        """"""
        pass  # Override in subclass to service receives and sends




    def exit(self):
        """"""
        self.peer.close()




# -*- encoding: utf-8 -*-
"""
hio.core.hier.hiering Module

Provides hierarchical action support




Syntax notes for Mine and Dock use in boxwork:
    M mine (Mine) memory non-durable attribute syntax
    D dock (Dock) durable attribute syntax
    with locally scope variables M ref for mine and D ref for dock.


    so need syntax does not heed to "quote" paths keys into the bags
    containers Mine dict subclasses with attribute support.

    M.root_dog.value  is equivalent to self.mine["root_dog"].value or
                                       self.mine[("root", "dog")].value

    So need term  "M.root_dog.value > 5" should compile directly and eval
    as long as M is in the locals() and M is a Mine instance.

    Likewise for
    D.root_dog.value where D is a Dock and root_dog is a key in the Dock.

    So no need to do substitutions or shorthand
    The hierarchy in the .mine/.dock is indicated by '_' separated keys
    The Box Boxer Actor names are forbidden from having '_" as an element
    with Renam regex test.



"""
from __future__ import annotations  # so type hints of classes get resolved later

import os
import sys
import time
import logging
import json
import signal
import re
import multiprocessing as mp
import functools

from collections import deque, namedtuple
from collections.abc import Iterable, Mapping, Callable
from typing import Any, Type
from dataclasses import dataclass, astuple, asdict, field


from ..tyming import Tymee
from ... import hioing
from ...hioing import Mixin, HierError
from ...help import isNonStringIterable, MapDom, modify, Renam, Mine


"""Contextage (namedtuple):
Action contexts for Acts

Fields:
   native (str): native context
   precon (str): precon context
   renter (str): renter context
   enter (str): enter context
   recur (str): recur context
   tail (str): tail context
   transit (str): transit context
   exit (str): exit context
   rexit (str): rexit context
"""
Contextage = namedtuple("Contextage", "native precon renter enter recur tail transit exit rexit")

Context = Contextage(native="native", precon="precon", renter="renter",
                     enter="enter", recur="recur", tail="tail",
                     transit="transit", exit="exit", rexit="rexit")


@dataclass
class WorkDom(MapDom):
    """WorkDom provides state for building boxwork by a boxer to be injected
    make methods of Boxer by workify wrapper.


    Attributes:
        box (Box | None): current box in box work. None if not yet a box
        over (Box | None): current over Box in box work. None if top level
        bxpre (str):  default box name prefix used to generate unique box name
                    relative to boxer.boxes
        bxidx (int): default box name index used to generate unique box name
                    relative to boxer.boxes
        acts (dict):  registry of ActBase subclasses by name (including aliases)
    """
    box: None | Box = None  # current box in boxwork. None if not yet any box
    over: None | Box = None  # current over box in boxwork. None if not yet any over
    bxpre: str = 'box'  # default box name prefix when name not provided
    bxidx: int = 0  # default box name index when name not provided
    acts: dict = field(default_factory=dict)  # registry of Acts by name & aliases
    context: str = Context.native  # current context



def actify(name, *, base=None, attrs=None):
    """Parametrized decorator that converts the decorated function func into
    .act method of new subclass of class base with .__name__ name. If not
    provided then uses Act as base. When provided base must be subclass of ActBase.
    Registers new subclass in ActBase.Registry. Then instantiates cls and returns
    instance.

    Returns:
        instance (cls): instance of new subclass

    Updates the class attributes of new subclass with attrs if any.

    Any callable usually function that is not already a subclass of ActBase can be
    converted to a subclass of ActBase and then registered in the Registry by its
    given name. Usually when defining a class simply make is a subclass of ActBase.
    But when given a function on can decorate it with actify so that a new
    subclass of ActBase is created and registered.

    Usage:
        @actify(name="Tact")
        def test(self, **kwa):  # signature for .act with **iops as **kwa
            assert kwa == self.iops
            return self.iops

        t = test(iops=dict(what=1), hello="hello", context=Context.recur)

    Notes:
    In Python, when a function is assigned as the value of a class attribute,
    the value of that attribute is automagically converted to a bound method of
    that class with injected self as first argument.

    class A():
        def a(self):
           print(self)

    a = A()
    a.a()
    <__main__.A object at 0x1059ffc50>

    def b(self):
        print(self)

    A.b = b
    a.b()
    <__main__.A object at 0x1059ffc50>


    """
    base = base if base is not None else ActBase
    if not issubclass(base, ActBase):
        raise hioing.HierError(f"Expected ActBase subclass got {base=}.")

    attrs = attrs if attrs is not None else {}
    # assign Act attrs
    attrs.update(dict(Registry=base.Registry, Instances=base.Instances, Index=0,
                      Names=() ))
    cls = type(name, (base, ), attrs)  # create new subclass of base of Act.
    cls.registerbyname()  # register subclass in cls.Registry by cls.__name__

    def decorator(func):
        if not isinstance(func, Callable):
            raise hioing.HierError(f"Expected Callable got {func=}.")
        cls.act = func  # assign as bound method .act with injected self.

        @functools.wraps(func)
        def inner(*pa, **kwa):
            return cls(*pa, **kwa)  # instantiate and return instance
        return inner
    return decorator


def registerone(cls):
    """Class Decorator to add cls as cls.Registry entry for itself keyed by its
    own .__name__. Need class decorator so that class object is already created
    by registration time when decorator is applied
    """
    name = cls.__name__
    if name in cls.Registry:
        raise hioing.HierError(f"Act by {name=} already registered.")
    cls.Registry[name] = cls
    return cls


def register(names=None):
    """Parametrized class Decorator to add cls as cls.Registry entry for itself
    keyed by its own .__name__ as well as being keyed by alises in names.
    A class decorator is necessary so that the class object is already created
    when decorator is applied.

    Parameters:
        names (None|str|Iterator): iterator of names as aliases besides class
            name to register class in Act registry and assigned to cls.Names

    """
    if not names:
        names = tuple()
    elif isinstance(names, str):
        names = (names, )  # make one tuple of str
    else:
        names = tuple(names)  # make tuple of iterable

    def decorator(cls):
        name = cls.__name__
        if name in cls.Registry:  # already registration for name
            if cls.Registry[name] is not cls:  # different cls at name
                raise hioing.HierError(f"Different Act by same {name=} already "
                                       f"registered.")
        else:  # not yet regisered under name
            cls.Registry[name] = cls
        cls.Names = names  # creates class specific attribute .Names
        for name in names:
            if name in cls.Registry:   # already registration for name
                if cls.Registry[name] is not cls:  # different cls at name
                    raise hioing.HierError(f"Different Act by same {name=} already "
                                           f"registered.")
            else:  # not yet registered under name
                cls.Registry[name] = cls
        return cls
    return decorator


@register()
class ActBase(Mixin):
    """Act Base Class. Callable with Registry of itself and its subclasses.

    Class Attributes:
        Registry (dict): subclass registry whose items are (name, cls) where:
                name is unique name for subclass
                cls is reference to class object
        Instances (dict): instance registry whose items are (name, instance) where:
                name is unique instance name and instance is instance reference
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.
        Names (tuple[str]): tuple of aliases (names) under which this subclas
                            appears in .Registry. Created by @register

    Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork

    Properties:
        name (str): unique name string of instance
        iops (dict): input-output-parameters for .act
        context (str): action context for .act

    Hidden
        ._name (str|None): unique name of instance
        ._iopts (dict): input-output-paramters for .act
        ._context (str): action context for .act

    """
    Registry = {}  # subclass registry
    Instances = {}  # instance name registry
    Index = 0  # naming index for default names of this subclasses instances
    #Names = ()  # tuple of aliases for this subclass created by @register


    @classmethod
    def registerbyname(cls, name=None):
        """Adds cls to cls.Registry entry by name. Raises HierError if already
        registered.

        Parameters:
            cls (Type[Act]): class to be registered
            name (None|str): key to register cls under.
                             When None then use cls.__name__
        """
        name = name if name is not None else cls.__name__
        if name in cls.Registry:  # already registration for name
            if cls.Registry[name] is not cls:  # different cls at name
                raise hioing.HierError(f"Different Act by same {name=} already "
                                           f"registered.")
        else:  # not yet registered under name
            cls.Registry[name] = cls
        return cls

    @classmethod
    def _clear(cls):
        """Clears Instances and and Index for testing purposes"""
        cls.Instances = {}
        cls.Index = 0


    @classmethod
    def _clearall(cls):
        """Clears Instances and and Index for testing purposes"""
        registry = ActBase.Registry.values()
        registry = set(registry)  # so only one copy, accounts for aliases
        for klas in registry:
            klas._clear()



    def __init__(self, *, name=None, iops=None, context=Context.enter,
                 mine=None, dock=None, **kwa):
        """Initialization method for instance.

        Parameters:
            name (str|None): unique name of this instance. When None then
                generate name from .Index
            iops (dict|None): input-output-parameters for .act. When None then
                set to empty dict.
            context (str): action context for act. default is "enter"
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork

        """
        super(ActBase, self).__init__(**kwa) # in case of MRO
        self.name = name  # set name property
        self._iops = dict(iops) if iops is not None else {}  # make copy
        self._context = context if context is not Context.native else Context.enter
        self.mine = mine if mine is not None else Mine()
        self.dock = dock   # stub fix later when have Dock class


    def __call__(self):
        """Make ActBase instance a callable object. run its .act method"""
        return self.act(**self.iops)


    def act(self, **iops):
        """Act called by Actor. Should override in subclass."""
        return iops  # for debugging


    @property
    def name(self):
        """Property getter for ._name

        Returns:
            name (str): unique identifier of instance
        """
        return self._name


    @name.setter
    def name(self, name=None):
        """Property setter for ._name

        Parameters:
            name (str|None): unique identifier of instance. When None generate
                unique name using .Index
        """
        while name is None or name in self.Instances:
                name = self.__class__.__name__ + str(self.Index)
                self.__class__.Index += 1   # do not shadow class attribute

        if not Renam.match(name):
            raise HierError(f"Invalid {name=}.")

        self.Instances[name] = self
        self._name = name


    @property
    def iops(self):
        """Property getter for ._iopts. Makes ._iopts read only

        Returns:
            iops (dict): input-output-parameters for .act
        """
        return self._iops


    @property
    def context(self):
        """Property getter for ._context. Makes ._context read only

        Returns:
            context (str): action context for .act
        """
        return self._context





# -*- encoding: utf-8 -*-
"""
hio.core.hier.boxing Module

Provides hierarchical box work support


"""
from __future__ import annotations  # so type hints of classes get resolved later

from collections.abc import Callable

from ..tyming import Tymee
from ...hioing import Mixin, HierError
from .hiering import Context, WorkDom, ActBase
from .acting import Act, Tract
from .bagging import Bag
from .needing import Need
from ...help import modify, Mine, Renam


contextDispatch = dict(precon="preacts",
                       renter="renacts",
                       enter="enacts",
                       recur="reacts",
                       tail="tacts",
                       transit="tracts",
                       exit="exacts",
                       rexit="rexacts")


class Box(Tymee):
    """Box Class for hierarchical action framework (boxwork) instances.
    Box instance holds reference to in-memory data mine shared by all the boxes in a
    given boxwork as well as its executing Boxer.
    Box instance holds references (links) to its over box and its under boxes.
    Box instance holds the acts to be executed in their context.

    Inherited Attributes, Properties
        see Tymee

    Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork
        over (Box | None): this box's over box instance or None
        unders (list[Box]): this box's under box instances or empty
                            zeroth entry is primary under
        preacts (list[act]): precon (pre-conditions for entry) context acts
        remacts (list[act]): remark re-enter mark subcontext acts (retained)
        renacts (list[act]): renter (re-enter) context acts  (retained)
        enmacts (list[act]): enmark enter mark subcontext acts
        enacts (list[act]):  enter context acts
        reacts (list[act]): recur context acts
        tacts (list[act]): tail context acts
        tracts (list[act]): transit context acts
        exacts (list[act]): exit context acts
        rexacts (list[act]): rexit (re-exit) context acts  (retained)

    Properties:
        name (str): unique identifier of instance
        pile (list[Box]): this box's pile of boxes generated by tracing .over up
                          and .unders[0] down if any. This is generated lazily.
                          To refresh call ._trace()
        trail (str): human friendly represetion of pile as delimited string of
                        box names from .pile. This is generated lazily.
                        To refresh call ._trace()

    Hidden:
        _name (str): unique identifier of instance
        _pile (list[Box] | None): pile of Boxes to which this box belongs.
                                  None means not yet traced.
        _spot (int | None): zero based offset into .pile of this box. This is
                            computed by ._trace
        _trail (int | None): human friendly represetion of pile as delimited
                             string of box names from .pile.
                            This is computed by ._trace
        _trace(): function to trace and update ._pile from .over and .unders[0]
                  and update ._spot and ._trail
        _next (Box | None): this box's next box if any lexically




    """
    def __init__(self, *, name='box', mine=None, dock=None, over=None, **kwa):
        """Initialize instance.

        Parameters:
            name (str): unique identifier of box
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork
            over (Box | None): this box's over box instance or None
        """
        super(Box, self).__init__(**kwa)
        self.name = name
        self.mine = mine if mine is not None else Mine()
        self.dock = dock   # stub for now until create Dock class
        self._pile = None  # force .trace on first access of .pile property
        self._spot = None  # zero based offset into .pile of this box
        self._trail = None  # delimited string representation of box names in .pile
        self.over = over  # over box
        self.unders = []  # list of under boxes,

        # acts by contexts
        self.preacts = []  # precon context list of pre-entry acts
        self.remacts = []  # re-enter mark subcontext list of re-mark acts
        self.renacts = []  # renter context list of re-enter acts (retained)
        self.enmacts = []  # enter mark subcontext list of en-mark acts
        self.enacts = []  # enter context list of enter acts
        self.reacts = []  # recur context list of recurring acts
        self.tacts = []  # tail context list of trailing acts
        self.tracts = []  # transit context list of transition acts
        self.exacts = []  # exit context list of exit acts
        self.rexacts = []  # rexit context list of re-exit acts (retained)

        #lexical context
        self._next = None  # next box lexically


    def __repr__(self):
        """Representation usable by eval()."""
        return (f"{self.__class__.__name__}(name='{self.name}')")

    def __str__(self):
        """Representation human friendly."""
        return (f"{self.__class__.__name__}({self.trail})")


    def _trace(self):
        """Trace pile and update .pile by tracing over up if any and unders[0]
        down if any.
        """
        pile = []
        over = self.over
        while over:
            pile.insert(0, over)
            over = over.over
        pile.append(self)
        self._spot = len(pile) - 1
        under = self.unders[0] if self.unders else None
        while under:
            pile.append(under)
            under = under.unders[0] if under.unders else None
        self._pile = pile

        up = "<".join(over.name for over in self._pile[:self._spot])
        dn = ">".join(under.name for under in self._pile[self._spot+1:])
        self._trail = up + "<" + self._name + ">" + dn


    @property
    def name(self):
        """Property getter for ._name

        Returns:
            name (str): unique identifier of instance
        """
        return self._name


    @name.setter
    def name(self, name):
        """Property setter for ._name

        Parameters:
            name (str): unique identifier of instance
        """
        if not Renam.match(name):
            raise HierError(f"Invalid {name=}.")
        self._name = name


    @property
    def pile(self):
        """Property getter for ._pile

        Returns:
            pile (list[Box]): this box's pile of boxes generated by tracing
                              .over up and .unders[0] down if any. This is
                              generated lazily to refresh call ._trace().
                              pile always includes self once traced.
        """
        if self._pile is None:
            self._trace()
        return self._pile

    @property
    def spot(self):
        """Property getter for ._spot

        Returns:
            spot (int): zero based offset of this box into its pile of boxes
                        generated by tracing .over up and .unders[0] down if any.
                        This is generated lazily. To refresh call ._trace().
                        Since pile always includes self, spot is always defined
                        once traced.
        """
        if self._spot is None:
            self._trace()
        return self._spot

    @property
    def trail(self):
        """Property getter for ._trail

        Returns:
            trail (str): human frieldly delimited string of box names from .pile.
                        This is generated lazily. To refresh call ._trace().
                        Since pile always includes self, trail is always defined
                        once traced.
        """
        if self._trail is None:
            self._trace()
        return self._trail





class Boxer(Tymee):
    """Boxer Class that executes hierarchical action framework (boxwork) instances.
    Boxer instance holds reference to in-memory data mine shared by all its boxes
    and other Boxers in a given boxwork.
    Box instance holds a reference to its first (beginning) box.
    Box instance holds references to all its boxes in dict keyed by box name.

    Inherited Attributes, Properties
        see Tymee

    Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork
        doer (Doer | None): doer running this boxer  (do we need this?)
        boxes (dict): all boxes mapping of (box name, box) pairs
        first (Box | None):  beginning box
        box (Box | None):  active box
        renters (list[Box]): boxes to re-enter on this run (boxes retained)
        enters (list[Box]): boxes to enter on this begin/run

    Properties:
        name (str): unique identifier of instance

    Hidden:
        _name (str): unique identifier of instance


    Cycle Context Order

    Init:

    all tymes in bags are None from init
    all tyme in marks are None



    tyme = start tyme (default ) 0.0
    Begin:

        First box assigned to active box
        actives = active box.pile
        .renters is empty
        .enters == actives
        exits is empty
        rexits is empty

        for each box in enters top down:
            precond:
                if not all true then done with boxwork
                    End

        for box in renters top down: (empty)
            remark renter mark subcontext
            renter
        for box in enters top down: (not empty)
            enmark enter mark subcontext  (mark tyme set to bag._tyme which is None)
            enter
        If complete:
            End boxwork boxer.want stop desire stop
        for box in actives top down:
            recur
            tail
            if transit need is true:  (short circuit recur of lower level boxes)
                compute renters enters exits rexits. save box.renters and box.enters
                for each box in enters (top down):
                    precon:
                        if not all true then do not proceed with transit return

                for box in exits bottom up:
                    exits
                for box in rexits bottom up:
                    rexits
                set new .active box
                break
        return from prep True == completed begin  (or yield if generator)


    tyme increment
    Run:
        for box in renters: (may be empty)
            remark renter mark subcontext
            renter
        for box in enters: (may be empty)
            enmark enter mark subcontext  (mark tyme set to bag._tyme which is None)
            enter
        If complete:
            End boxwork boxer.want stop desire stop
        for box in actives top down:
            recur
            tail
            if transit need is true:   (short circuit recur of lower level boxes)
                compute renters enters exits rexits,.
                    precon:
                        if not all true then do not proceed with transit return

                for box in exits bottom up:
                    exit
                for box in rexits bottom up:
                    rexit
                set new .active box
                break
        return from run False == not done continue  (or if generator yield)

    """
    def __init__(self, *, name='boxer', mine=None, dock=None, doer=None, **kwa):
        """Initialize instance.

        Parameters:
            name (str): unique identifier of box
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork
            doer (Doer | None): Doer running this Boxer doe we need this?


        """
        super(Boxer, self).__init__(**kwa)
        self.name = name
        self.mine = mine if mine is not None else Mine()
        self.dock = dock  # stub until create Dock class
        self.doer = doer

        self.boxes = {}
        self.first = None  # box to start in
        self.box = None  # current active box  whose pile is active pile
        self.renters = []  # list of re-enter boxes for this run (boxes retained)
        self.enters = []  # list of enter boxes for this begin/run


    @property
    def name(self):
        """Property getter for ._name

        Returns:
            name (str): unique identifier of instance
        """
        return self._name


    @name.setter
    def name(self, name):
        """Property setter for ._name

        Parameters:
            name (str): unique identifier of instance
        """
        if not Renam.match(name):
            raise HierError(f"Invalid {name=}.")
        self._name = name


    def begin(self):
        """Prepare and execute first pass
        """
        if not self.first:
            self.first = list(self.boxes.values())[0]  # first box in boxes is default first
        self.box = self.first
        self.renters = []  # first pass no rentry of any boxes
        self.enters = self.box.pile
        if not self.precon():  # uses .enters preconditions for entry not satisfied
            # do end stuff since no entry yet then no exit
            self.box = None  # no active box anymore
            return False  # signal end beginning did not complete

        #self.renter()  # uses saved .renters
        #self.renters = []  # re-entry completed
        self.enter()  # uses saved .enters to en-mark and enter
        self.enters = []  # entry completed
        # check for End here if so .end()
        for box in self.box.pile:
            for react in box.reacts:   # recur context top down
                react()
            for tact in box.tacts:   # trail context top down
                tact()
            if self.endial():  # actioned desire to end
                self.end()  # exits all active boxes in self.box.pile
                self.box = None  # no active box
                return True  # beginning completed already

            for tract in box.tracts:  # transit context top down
                if tract():  # transition condition satisfied
                    exits, enters, renters, rexits = self.exen(box, tract.dest)
                    if not self.precon(enters):  # transit not satisfied
                        continue  # keep trying
                    self.exit(exits)  # exit bottom up
                    self.rexit(rexits)  # rexit bottom up
                    self.renters = renters  # save for next pass
                    self.enters = enters  # save for next pass
                    self.box = tract.dest  # set new active box
                    break

        return True  # successfully completed beginning


    def run(self):
        """Execute another pass
        """
        self.renter()  # uses saved .renters to re-mark and re-enter
        self.renters = []  # re-entry completed so make empty
        self.enter()  # uses saved .enters to en-mark and enter
        self.enters = []  # entry completed so make empty

        for box in self.box.pile:  # top down
            for react in box.reacts:   # recur context top down
                react()
            for tact in box.tacts:   # trail context top down
                tact()

            if self.endial():  # actioned desire to end
                self.end()  # exits all active boxes in self.box.pile
                self.box = None  # no active box
                return True  # beginning completed already

            for tract in box.tracts:  # transit context top down
                if dest := tract():  # transition condition satisfied
                    exits, enters, renters, rexits = self.exen(box, dest)
                    if not self.precon(enters):  # transit not satisfied
                        continue  # keep trying
                    self.exit(exits)  # exit bottom up
                    self.rexit(rexits)  # rexit bottom up  (boxes retained)
                    self.renters = renters  # save for next pass
                    self.enters = enters  # save for next pass
                    self.box = dest  # set new active box
                    return False  # transition so stop iteration over pile

        return False  # continue running not done


    def end(self):
        """Exit all active boxes.

        """
        self.exit(self.box.pile)  # exit all active boxes


    def endial(self):
        """Check for desire to end execution and return True otherwise False
        End condition if bag alue at mine._boxer_name_end.value == True

        Returns:
            end (bool): True means end condition satisfied
                        False otherwise

        if keys not in self.mine:
            self.mine[keys] = Bag()  # create bag at end default value = None
        """
        keys = ("", "boxer", self.name, "end")  # _boxer_boxername_end
        if keys in self.mine and self.mine[keys].value:
            return True

        return False


    def precon(self, enters=None):
        """Evaluate preconditions for entry of boxes in enters in top down order

        Parameters:
            enters (None|list[Box]): boxes to be entered if precons are satisfied
                                default when None is .enters

        Returns:
            met (bool): True means all preconditions are satisfied for enters.
                        False otherwise
                    When no preconditions then returns True.

        """
        enters = enters if enters is not None else self.enters
        met = True
        for box in self.enters:
            for preact in box.preacts:
                if not preact():
                    met = False
                    break
            if not met:
                break
        return met


    def renter(self):
        """Action re-mark (remacts) and re-enter (renacts) acts of boxes in
        .renters in top down order. Boxes retained in hierarchical state.
        """
        for box in self.renters:
            for remact in box.remacts:
                remact()
            for renact in box.renacts:
                renact()


    def enter(self):
        """Action e-mark (emacts) and enter (enacts) acts of boxes in .enters
        in top down order
        """
        for box in self.enters:
            for enmact in box.enmacts:
                enmact()
            for enact in box.enacts:
                enact()


    def exit(self, exits):
        """Action exacts of boxes in exits in bottom up order.

        Parameters:
            exits (None|list[Box]): boxes to be exited in bottom up order

        """
        for box in exits:
            for exact in box.exacts:
                exact()


    def rexit(self, rexits):
        """Action rexacts of boxes in rexits (re-exits) in bottom up order.
        Boxes retained in hierarchical state.

        Parameters:
            rexits (None|list[Box]): boxes to be re-exited in bottom up order
        """
        for box in rexits:
            for rexact in box.rexacts:
                rexact()


    def resolve(self):
        """Resolve both over box names and tract dest box names into boxes for
        all boxes in .boxes

        """
        for name, box in self.boxes.items():
            if isinstance(box.over, str):
                try:
                    over = self.boxes[over]  # resolve
                except KeyError as ex:
                    raise HierError(f"Unresolvable over box name={over} for"
                                           f"box {name=}.") from ex
                box.over = over  # resolve over as a box
                box.over.unders.append(box)  # add box to its over.unders list

            for tract in box.tracts:
                if isinstance(tract.dest, str):
                    if tract.dest == 'next':  # next
                        if not box._next:
                            HierError(f"Unresolvable dest 'next' for tract in "
                                      f"box{name=}")
                        dest = box._next
                    else:
                        try:
                            dest = self.boxes[tract.dest]  # resolve
                        except KeyError as ex:
                            raise HierError(f"Unresolvable dest box={tract.dest}"
                                f" for tract in box{name=}") from ex

                    tract.dest = dest  # resolve dest as box

        if isinstance(self.first, str):  # resolve first box
            try:
                self.first = self.boxes[self.first]  # resolve
            except KeyError as ex:
                raise HierError(f"Unresolvable over box name={self.first} for"
                                       f"boxer {self.name}.") from ex


    def make(self, fun):
        """Make box work for this boxer from function fun
        Parameters:
            fun (function):  employs be, do, on, go maker functions injected
                             works (boxwork state vars)

        def fun(bx):


        Injects works as WorkDom dataclass instance whose attributes are used to
        construct boxwork. WorkDom attributes include
            box (Box|None): current box in box work. None if not yet a box
            over (Box|None): current over Box in box work. None if top level
            bxpre (str): default name prefix used to generate unique box
                name relative to boxer.boxes
            bxidx (int): default box index used to generate unique box
                name relative to boxer.boxes


        """
        works = WorkDom()  # standard defaults
        works.acts = ActBase.Registry
        bx = modify(mods=works)(self.bx)
        on = modify(mods=works)(self.on)
        go = modify(mods=works)(self.go)
        do = modify(mods=works)(self.do)
        fun(bx=bx, on=on, go=go, do=do)  # calling fun will build boxer.boxes
        self.resolve()
        return works  # for debugging analysis



    def bx(self, name: None|str=None, over: None|str|Box="",
                *, mods: WorkDom|None=None)->Box:
        """Make a box and add to box work

        Parameters:
            name (None | str): when None then create name from bepre and beidx
                               items in works.
                               if non-empty string then use provided
                               otherwise raise exception

            over (None | str | Box): over box for new box.
                                    when str then name of new over box
                                    when box then actual over box
                                    when None then no over box (top level)
                                    when empty then same level use _over

            mods (None | WorkDom):  state variables used to construct box work
                None is just to allow definition as keyword arg. Assumes in
                actual usage that mods is always provided as WorkDom instance of
                form:

                    box (Box|None): current box in box work. None if not yet a box
                    over (Box|None): current over Box in box work. None if top level
                    bxpre (str): default name prefix used to generate unique box
                        name relative to boxer.boxes
                    bxidx (int): default box index used to generate unique box
                        name relative to boxer.boxes



        """
        m = mods  # alias more compact

        if not name:  # empty or None
            if name is None:
                name = m.bxpre + str(m.bxidx)
                m.bxidx += 1
                while name in self.boxes:
                    name = m.bxpre + str(m.bxidx)
                    m.bxidx += 1

            else:
                raise HierError(f"Missing name.")

        if name in self.boxes:  # duplicate name
            raise HierError(f"Non-unique box {name=}.")

        if over is not None:  # not at top level
            if isinstance(over, str):
                if not over:  # empty string
                    over = m.over  # same level
                else:  # resolvable string
                    try:
                        over = self.boxes[over]  # resolve
                    except KeyError as ex:
                        raise HierError(f"Under box={name} defined before"
                                               f"its {over=}.") from ex

            elif over.name not in self.boxes:  # stray over box
                self.boxes[over.name] = over  # add to boxes

        box = Box(name=name, over=over, mine=self.mine, tymth=self.tymth)
        self.boxes[box.name] = box  # update box work
        if box.over is not None:  # not at top level
            box.over.unders.append(box)  # add to over.unders list

        m.over = over  # update current level
        if m.box:  # update last boxes lexical ._next to this box
            m.box._next = box
        m.box = box  # update current box
        return box


    def on(self, cond: None|str=None, key: None|str=None, expr: None|str=None,
                 *, mods: WorkDom|None=None, **kwa)->Need:
        """Make a special Need and return it.
        Used for special needs for tracts and also for beacts (before enter)

        Returns:
            need (Need):  newly created special need

        Parameters:
            cond (None|str): special need condition to be satisfied. This is
                resolved in evalable boolean expression.
                When None then ignore
                When str then special need condition to be resolved into evalable
                boolean expression

            key (None|str): key to mine item ref for special need cond when
                applicable, i.e. cond is with respect to mine at key that is
                not predetermined solely by cond. Otherwise None.
                When None use default for cond
                When str then resolve key to mine at key


            expr (None|str): evalable boolean expression as additional constraint(s)
                ANDed with result of cond.
                When None or empty then ignore
                When str then evalable python boolean expression to be ANDed with
                    the result of cond resolution.

            mods (None | WorkDom):  state variables used to construct box work
                None is just to allow definition as keyword arg. Assumes in
                actual usage that mods is always provided as WorkDom instance of
                form:

                    box (Box|None): current box in box work. None if not yet a box
                    over (Box|None): current over Box in box work. None if top level
                    bxpre (str): default name prefix used to generate unique box
                        name relative to boxer.boxes
                    bxidx (int): default box index used to generate unique box
                        name relative to boxer.boxes



        """
        m = mods  # alias more compact

        _expr = None

        if not cond:
            if not expr:
                cond = "updated"  # default
            else:  # no cond but with expr
                _expr = expr  # use expr instead of resolved cond
                expr = None  # can't have both _expr and expr same below

        if not _expr:  # cond above so need to resolve cond into _expr
            if cond == "updated":
                _expr = "True"
            else:
                pass  # raise error since must have valid _expr after here

        # now _expr is valid

        if expr:  # both resolved cond as _expr and expr so AND together
            _expr = "(" + _expr + ") and (" + expr + ")"



        need = Need(expr=_expr, mine=self.mine, dock=self.dock)


        return need




    def go(self, dest: None|str=None, expr: None|str=None,
                 *, mods: WorkDom|None=None, **kwa)->Tract:
        """Make a Tract and add it to the tracts context of the current box.

        Returns:
            tract (Tract):  newly created tract

        Parameters:
            dest (None|str|Box): destination box its name for transition.
                When None use next box if any
                When str then resolve name to box if possible else save for
                    later resolution
                When Box instance that already resolved

            expr (None|str): evalable boolean expression for transition to dest.
                When None then conditional always True. Always transit.
                When str then evalable python boolean expression to be
                    resolved into a Need instance for eval at run time

            mods (None | WorkDom):  state variables used to construct box work
                None is just to allow definition as keyword arg. Assumes in
                actual usage that mods is always provided as WorkDom instance of
                form:

                    box (Box|None): current box in box work. None if not yet a box
                    over (Box|None): current over Box in box work. None if top level
                    bxpre (str): default name prefix used to generate unique box
                        name relative to boxer.boxes
                    bxidx (int): default box index used to generate unique box
                        name relative to boxer.boxes



        """
        m = mods  # alias more compact

        if not dest or dest in ('next', 'Next', 'NEXT'):  # empty or None or next
            if m.box._next:
                dest = m.box._next
            else:
                dest = 'next'  # to be resolved later
        elif isinstance(dest, str):
            if not Renam.match(dest):
                raise HierError(f"Invalid {dest=}.")
            if dest in self.boxes:
                dest = self.boxes[dest]

        need = Need(expr=expr, mine=self.mine, dock=self.dock)

        tract = Tract(dest=dest, need=need)
        m.box.tracts.append(tract)
        return tract



    def do(self, deed: None|str|Callable=None, *, name: str|None=None,
           mods: WorkDom|None=None, **iops)->str:
        """Make an act and add to box work

        Parameters:
            deed (None|str|Callable): When str name of class in ActBase registry.
                    When None use Act with default lambda and iops as parameters.
                    When Callable use Act with iops as parameters.


            mods (None | WorkDom):  state variables used to construct box work
                None is just to allow definition as keyword arg. Assumes in
                actual usage that mods is always provided as WorkDom instance.

            iops (dict): input-output-parms for Act

        """
        m = mods  # alias more compact

        parms = dict(name=name, mine=self.mine, dock=self.dock)
        if m.context != Context.native:
            parms.update(context=m.context)  # override default context for klas

        iops = dict(_boxer=self.name, _box=m.box.name, **iops)
        parms.update(iops=iops)


        deed = deed if deed is not None else "Act"

        if isinstance(deed, str):
            try:
                klas = m.acts[deed]
            except KeyError as ex:
                raise HierError(f"Unregistered deed='{deed}'") from ex

            act = klas(**parms)

        elif callable(deed):
            act = Act(deed=deed, **parms)

        else:
            raise HierError(f"Invalid {deed=}")

        context = act.context  # act init may override passed in context

        try:
            getattr(m.box, contextDispatch[context]).append(act)
        except (KeyError, AttributeError) as ex:
            raise HierError("Unrecognized context='{context}'") from ex


        return act


    @staticmethod
    def exen(near,far):
        """Computes the relative differences (uncommon  and common parts) between
        the box pile lists nears passed in and fars from box far.pile

        Parameters:
            near (Box): near box giving nears =near.pile in top down order
            far (Box): far box giving fars = far.pile in top down order.

        Assumes piles nears and fars are in top down order

        Returns:
            quadruple (tuple[list]): quadruple of lists of form:
                (exits, enters, renters, rexits) where:
                exits is list of uncommon boxes in nears but not in fars to be exited.
                    Reversed to bottom up order.
                enters is list of uncommon boxes in fars but not in nears to be entered
                rexits is list of common boxes in both nears and fars to be re-exited
                    Reversed to bottom up order. These are boxes retained in pile.
                renters is list of common boxes in both nears and fars to be re-entered
                    These are boxes retained in pile.
                The sets of boxes in rexits and renters are the same set but rexits
                is reversed to bottum up order. These are boxes retained in the
                pile before and after the transition. This is where common exit/enter
                actions for the non-common boxes can be actioned non-redundantly.


        Supports forced reentry transitions when far is in nears. This means fars
            == nears. In this case:
            The common part of nears/fars from far down is force exited
            The common part of nears/fars from far down is force entered

        When far in nears then forced entry at far so far is nears[i]
        catches that case for forced entry at some far in nears. Since
        far is in fars, then when far == nears[i] then fars == nears.

        Since a given box's pile is always traced up via its .over if any and down via
        its primary under i.e. .unders[0] if any, when far is in nears the anything
        below far is same in both fars and nears.

        Otherwise when far not in nears then i where fars[i] is not nears[i]
        indicates first box where fars down and nears down is uncommon i.e. the pile
        tree branches at i. This is the normal non-forced entry case for transition.

        Two different topologies are accounted for with this code.
        Recall that python slice of list is zero based where:
           fars[i] not in fars[:i] and fars[i] in fars[i:]
           nears[i] not in nears[:i] and nears[i] in nears[i:]
           this means fars[:0] == nears[:0] == [] empty list

        1.0 near and far in same tree either on same branch or different branches
            1.1 on same branch forced entry where nears == fars so far in nears.
               Walk down from shared root to find where far is nears[i]. Boxes above
               far given by fars[:i] == nears[:i] are re-exit re-enter set of boxes.
               Boxes at far and below are forced exit entry.
            1.2 on different branch to walk down from root until find fork where
               fars[i] is not nears[i]. So fars[:i] == nears[:i] above fork at i,
               and are re-exit and re-enter set of boxes. Boxes at i and below in
               nears are exit and boxes at i and below in fars are enter
        2.0 near and far not in same tree. In this case top of nears at nears[0] is
            not top of fars ar fars[0] i.e. different tree roots, far[0] != near[0]
            and fars[:0] == nears[:0] = [] means empty re-exits and re-enters and
            all nears are exit and all fars are entry.

        """
        nears = near.pile  # top down order
        fars = far.pile  # top down order
        l = min(len(nears), len(fars))  # l >= 1 since far in fars & near in nears
        for i in range(l):  # start at the top of both nears and fars
            if (far is nears[i]) or (fars[i] is not nears[i]): #first effective uncommon member
                # (exits, enters, rexits, renters)
                return (list(reversed(nears[i:])), fars[i:],
                        list(reversed(nears[:i])), fars[:i])




class Maker(Mixin):
    """Maker Class makes boxworks of Boxer and Box instances.
    Holds reference to in-memory mine shared by all boxes in boxwork
    Holds reference to current Boxer and Box being built

    ****Placeholder for now. Future to be able to make multiple boxers from
    single fun or in multiple iterations making.****

    Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork
        boxer (Boxer | None): current boxer
        box (Box | None): cureent box

    Properties:
        name (str): unique identifier of instance

    Hidden:
        _name (str): unique identifier of instance

    """
    def __init__(self, *, name='maker', mine=None, dock=None, **kwa):
        """Initialize instance.

        Parameters:
            name (str): unique identifier of instance
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork


        """
        super(Maker, self).__init__(**kwa)
        self.name = name
        self.mine = mine if mine is not None else Mine()
        self.dock = dock  # stub until create Dock class
        self.boxer = None
        self.box = None

    @property
    def name(self):
        """Property getter for ._name

        Returns:
            name (str): unique identifier of instance
        """
        return self._name


    @name.setter
    def name(self, name):
        """Property setter for ._name

        Parameters:
            name (str): unique identifier of instance
        """
        if not Renam.match(name):
            raise HierError(f"Invalid {name=}.")

        self._name = name

    def make(self, fun, mine=None, boxes=None):
        """Make box work from function fun
        Parameters:
            fun (function):  employs be, do, on, go maker functions with
                              globals
            bags (None|Mine):  shared data Mine for all made Boxers
            boxes (None|dict): shared boxes map



        """

        # bags, boxes, and boxers can be referenced by fun in its nonlocal
        # enclosing scope. collections references so do not need to be global
        mine = mine if mine is not None else Mine()  # create new if not provided
        boxes = boxes if boxes is not None else {}  # create new if not provided
        boxers = []  # list of made boxers

        # create a default boxer
        boxer = Boxer(name='boxer', mine=mine, boxes=boxes)
        boxers.append(boxer)

        fun()



# -*- encoding: utf-8 -*-
"""
hio package
"""

__version__ = '0.7.6'  # also change in setup.py

from .hioing import (Mixin, HioError, SizeError, ValidationError, VersionError,
                     HierError)



# -*- encoding: utf-8 -*-
"""hio.core.hier.bagging Module

Provides dom support for items in Mine or Dock for shared data store support

"""
from __future__ import annotations  # so type hints of classes get resolved later

from collections import deque, namedtuple
from collections.abc import Iterable, Mapping, Callable
from typing import Any, Type
from dataclasses import dataclass, astuple, asdict, field

from ...help import RawDom

@dataclass
class BagDom(RawDom):
    """BagDom is base class for Bag subclasses with common fields.


    Attributes:
        _tyme (None|Float): tyme of last update of bag during prep/run time
                            None means not yet updated during prep/run time.
                            bag may be initialized during init time but
                            does not get a non-None tyme until its first update
                            during prep/run time.
    """
    _tyme: None|float = None  # tyme of last update


@dataclass
class Bag(BagDom):
    """Bag is simple BagDom sublclass with generic value field.


    Inherited Attributes:
        _tyme (None|Float): tyme of last update of bag during prep/run time
                            None means not yet updated during prep/run time.
                            bag may be initialized during init time but
                            does not get a non-None tyme until its first update
                            during prep/run time.

    Attributes:
       value (Any):
    """
    value: Any = None  # generic value



# -*- encoding: utf-8 -*-
"""
hio.base.hier Package
"""
from .hiering import Context, WorkDom, actify, register, ActBase
from .boxing import Box, Boxer, Maker
from .acting import Act, Tract, EndAct
from .needing import Need
from .bagging import BagDom, Bag



# -*- encoding: utf-8 -*-
"""
hio.core.hier.needing Module

Provides hierarchical action support

"""
from __future__ import annotations  # so type hints of classes get resolved later

from collections.abc import Callable, Iterable
from collections import namedtuple
from types import CodeType

from ... import hioing
from ...hioing import Mixin, HierError
from ...help import Mine


class Need(Mixin):
    """Need is conditional callable class whose callable returns a boolean.
    The calling it evaluates a need expression. May be used as the transition
    condition of a Gact.

    Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork

    Properties:
        expr (str): evalable boolean expression.
        compiled (bool): True means ._code holds compiled ._expr
                         False means not yet compiled


    Hidden:
        _expr (str): evalable boolean expression.
        _code (None|CodeType): compiled evalable boolean expression .expr
            None means not yet compiled from .expr



    Compilation Notes:
        c = compile("True", '<string>', 'eval')
        c
        <code object <module> at 0x1042f2250, file "<string>", line 1>
        type(c)
        <class 'code'>
        import types
        isinstance(c, types.CodeType)
        True

        import pickle
        s = pickle.dumps(c)
        builtins.TypeError: cannot pickle code objects

        eval(c)
        True

        So  the need returned by an on() call must keep around the string
        representation for multiprocessing because the code object itself is not
        pickleable and must be recompiled inside the child process.
        Also having the string rep around can be helpful in debugging if define
        __repr__ and __str__ for need objects.

    Expr syntax notes:
        M (Mine) memory (non-durable) and D (Dock) durable support attribute syntax
        with locally scope variables M ref for mine and D ref for dock.


        so need syntax does not heed to "quote" paths keys into the bags
        containers Mine dict subclasses with attribute support.

        M.root_dog.value  is equivalent to self.mine["root_dog"].value or
                                           self.mine[("root", "dog")].value

        So need term  "M.root_dog.value > 5" should compile directly and eval
        as long as M is in the locals() and M is a Mine instance.

        Likewise for
        D.root_dog.value where D is a Dock and root_dog is a key in the Dock.

        So no need to do substitutions or shorthand
        The hierarchy in the .mine and .dock is indicated by '_' separated keys
        The Box Boxer Actor names are forbidden from having '_" as an element
        with Renam regex test.

    """


    def __init__(self,  *, expr='True', mine=None, dock=None, **kwa):
        """Initialization method for instance.

        Parameters:
            expr (str): evalable boolean expression.
                        if empty or None then use default = 'True'
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork


        """
        super(Need, self).__init__(**kwa)
        self.expr = expr
        self.mine = mine if mine is not None else Mine()
        self.dock = dock   # stub fix later when have Dock class


    def __call__(self):
        """Make Need instance a callable object."""
        if not self.compiled:  # not yet compiled so lazy
            self.compile()  # first time only recompile
        M = self.mine  # ensure M is in locals() for eval
        D = self.dock  # ensure D is in locals() for eval
        return eval(self._code)


    @property
    def expr(self):
        """Property getter for ._expr

        Returns:
            expr (str): evalable boolean expression.
        """
        return self._expr


    @expr.setter
    def expr(self, expr=None):
        """Property setter for ._expr

        Parameters:
            expr (str): evalable boolean expression.
        """
        self._expr = expr if expr else 'True'
        self._code = None  # force lazy recompilation


    @property
    def compiled(self):
        """Property compiled

        Returns:
            compiled (bool): True means ._code holds compiled ._expr
                             False means not yet compiled
        """
        return True if self._code is not None else False


    def compile(self):
        """Compile evable boolean expression str ._expr into compiled code
        object ._code to be evaluated at run time.
        Because code objects are not pickleable the compilation must happen
        at prep (enter) time not init time.
        """
        self._code = compile(self.expr, '<string>', 'eval')




# -*- encoding: utf-8 -*-
"""
hio.help package

"""

# Setup module global ogler as package logger factory. This must be done on
# import to ensure global is defined so all modules in package have access to
# logggers via help.ogler.getLoggers().
# May always change level and reopen log file if need be.

from . import ogling

# initialize global ogler at hio.help.ogler always instantiated by default
ogler = ogling.initOgler(prefix='hio')  # init only runs  once on import

from .helping import isNonStringIterable, isNonStringSequence, isIterator, Reat
from .decking import Deck
from .hicting import Hict, Mict
from .timing import Timer, MonoTimer, TimerError, RetroTimerError
from .naming import Namer
from .doming import RawDom, MapDom, IceMapDom, modify, modize
from .mining import Renam, Mine




# -*- encoding: utf-8 -*-
"""
Demo web server for static files for client side web app
"""
import sys
import os
import time
import mimetypes

import falcon

from hio import help
from hio.base import tyming
from hio.core import http
from hio.core.http import serving
from hio.demo.web import demoing

logger = help.ogler.getLogger()



def run():
    """
    Use hio http server
    """
    tymist = tyming.Tymist(tyme=0.0)

    app = falcon.App()  # falcon.API instances are callable WSGI apps

    WEB_DIR_PATH = os.path.dirname(
                    os.path.abspath(
                        sys.modules.get(__name__).__file__))
    STATIC_DIR_PATH = os.path.join(WEB_DIR_PATH, 'static')


    sink = serving.StaticSink(staticDirPath=STATIC_DIR_PATH)
    app.add_sink(sink, prefix=sink.DefaultStaticSinkBasePath)

    try:
        server = http.Server(name='test',
                             port=8080,
                             tymeout=0.5,
                             app=app,
                             tymth=tymist.tymen())
        server.reopen()


        while True:
            try:
                server.service()
                time.sleep(0.0625)
                tymist.tick(tock=0.0625)

            except KeyboardInterrupt:  # use CNTL-C to shutdown from shell
                break

    finally:
        server.close()



if __name__ == '__main__':
    run()



# -*- encoding: utf-8 -*-
"""hio.help.packing Module

Packs and unpacks bit fields using format string whereas struct in std lib only
packs and unpacks byte fields.

"""



def packify(fmt=u'8', fields=[0x00], size=None, reverse=False):
    """
    Packs fields sequence of bit fields into bytearray of size bytes using fmt string.
    Each white space separated field of fmt is the length of the associated bit field
    If not provided size is the least integer number of bytes that hold the fmt.
    If reverse is true reverse the order of the bytes in the byte array before
    returning. This is useful for converting between bigendian and littleendian.

    Assumes unsigned fields values.
    Assumes network big endian so first fields element is high order bits.
    Each field in format string is number of bits for the associated bit field
    Fields with length of 1 are treated as has having boolean truthy field values
       that is,   nonzero is True and packs as a 1
    for 2+ length bit fields the field element is truncated to the number of
       low order bits in the bit field
    if sum of number of bits in fmt less than size bytes then the last byte in
       the bytearray is right zero padded
    if sum of number of bits in fmt greater than size bytes returns exception
    to pad just use 0 value in source field.
    example
    packify("1 3 2 2", (True, 4, 0, 3)). returns bytearry([0xc3])
    """
    tbfl = sum((int(x) for x in fmt.split()))
    if size is None:
        size = (tbfl // 8) + 1 if tbfl % 8 else tbfl // 8

    if not (0 <= tbfl <= (size * 8)):
        raise ValueError("Total bit field lengths in fmt not in [0, {0}]".format(size * 8))

    n = 0
    bfp = 8 * size  # starting bit field position
    bu = 0  # bits used

    for i, bfmt in enumerate(fmt.split()):
        bits = 0x0
        bfl = int(bfmt)
        bu += bfl

        if bfl == 1:
            bits = 0x1 if fields[i] else 0x0

        else:
            bits = fields[i] & (2**bfl - 1)  # bit-and mask out high order bits

        bits <<= (bfp - bfl) #shift left to bit position less bit field size

        n |= bits  # bit-or in bits
        bfp -= bfl #adjust bit field position for next element

    return bytify(n=n, size=size, reverse=reverse, strict=True)  # use int.to_bytes


def packifyInto(b, fmt=u'8', fields=[0x00], size=None, offset=0, reverse=False):
    """
    Packs fields sequence of bit fields using fmt string into bytearray b
    starting at offset and packing into size bytes
    Each white space separated field of fmt is the length of the associated bit field
    If not provided size is the least integer number of bytes that hold the fmt.
    Extends the length of b to accomodate size after offset if not enough.
    Returns actual size of portion packed into.
    The default assumes big endian.
    If reverse is True then reverses the byte order before extending. Useful for
    little endian.

    Assumes unsigned fields values.
    Assumes network big endian so first fields element is high order bits.
    Each field in format string is number of bits for the associated bit field
    Fields with length of 1 are treated as has having boolean truthy field values
       that is,   nonzero is True and packs as a 1
    for 2+ length bit fields the field element is truncated
    to the number of low order bits in the bit field
    if sum of number of bits in fmt less than size bytes then the last byte in
       the bytearray is right zero padded
    if sum of number of bits in fmt greater than size bytes returns exception
    to pad just use 0 value in source field.
    example
    packify("1 3 2 2", (True, 4, 0, 3)). returns bytearry([0xc3])
    """
    tbfl = sum((int(x) for x in fmt.split()))
    if size is None:
        size = (tbfl // 8) + 1 if tbfl % 8 else tbfl // 8

    if not (0 <= tbfl <= (size * 8)):
        raise ValueError("Total bit field lengths in fmt not in [0, {0}]".format(size * 8))

    if len(b) < (offset + size):
        b.extend([0x00]*(offset + size - len(b)))

    n = 0
    bfp = 8 * size  # starting bit field position
    bu = 0  # bits used

    for i, bfmt in enumerate(fmt.split()):
        bits = 0x0
        bfl = int(bfmt)
        bu += bfl

        if bfl == 1:
            bits = 0x1 if fields[i] else 0x0
        else:
            bits = fields[i] & (2**bfl - 1)  # bit-and mask out high order bits

        bits <<= (bfp - bfl) #shift left to bit position less bit field size

        n |= bits  # bit-or in bits
        bfp -= bfl #adjust bit field position for next element

    bp = bytify(n=n, size=size, reverse=reverse, strict=True) # use int.to_bytes

    b[offset:offset + len(bp)] = bp

    return size


def unpackify(fmt=u'1 1 1 1 1 1 1 1',
              b=bytearray([0x00]),
              boolean=False,
              size=None,
              reverse=False):
    """
    Returns tuple of unsigned int bit field values that are unpacked from the
    bytearray b according to fmt string. b maybe an integer iterator
    If not provided size is the least integer number of bytes that hold the fmt.
    The default assumes big endian.
    If reverse is True then reverse the byte order of b before unpackifing. This
    is useful for little endian.

    Each white space separated field of fmt is the length of the associated bit field.
    returns unsigned fields values.

    Assumes network big endian so first fmt is high order bits.
    Format string is number of bits per bit field
    If boolean parameter is True then return boolean values for
       bit fields of length 1

    if sum of number of bits in fmt less than 8 * size) then remaining
    bits are returned as additional field in result.

    if sum of number of bits in fmt greater 8 * len(b) returns exception

    example:
    unpackify(u"1 3 2 2", bytearray([0xc3]), False) returns (1, 4, 0, 3)
    unpackify(u"1 3 2 2", 0xc3, True) returns (True, 4, 0, 3)
    """
    b = bytearray(b)
    if reverse:
        b.reverse()

    tbfl = sum((int(x) for x in fmt.split()))
    if size is None:
        size = (tbfl // 8) + 1 if tbfl % 8 else tbfl // 8

    if not (0 <= tbfl <= (size * 8)):
        raise ValueError("Total bit field lengths in fmt not in [0, {0}]".format(size * 8))

    b = b[:size]
    fields = []  # list of bit fields
    bfp = 8 * size  # bit field position
    bu = 0  # bits used
    n = unbytify(b)  # unsigned int equivalent of b  # use int.from_bytes instead

    for i, bfmt in enumerate(fmt.split()):
        bfl = int(bfmt)
        bu += bfl

        mask = (2**bfl - 1) << (bfp - bfl)  # make mask
        bits = n & mask  # mask off other bits
        bits >>= (bfp - bfl)  # right shift to low order bits
        if bfl == 1 and boolean: #convert to boolean
            bits = True if bits else False

        fields.append(bits) #assign to fields list
        bfp -= bfl #adjust bit field position for next element

    if bfp != 0:  # remaining bits
        bfl = bfp
        mask = (2**bfl - 1) # make mask
        bits = n & mask  # mask off other bits
        if bfl == 1 and boolean: #convert to boolean
            bits = True if bits else False

        fields.append(bits) #assign to fields list
    return tuple(fields) #convert to tuple




# -*- encoding: utf-8 -*-
"""hio.core.hier.acting Module

Provides hierarchical action support

"""
from __future__ import annotations  # so type hints of classes get resolved later

from collections.abc import Callable
from collections import namedtuple

from ... import hioing
from ...hioing import Mixin, HierError
from ...help import Mine, Renam
from .hiering import Context, ActBase, register
from .needing import Need
from .bagging import Bag
from . import boxing




@register()
class Act(ActBase):
    """Act for do verb deeds as callables. At make (compile) time any callable
    that is available in the scope of the do verb in the boxer.make method
    can be passed in as the deed parameter and will be executed with ,iops as
    its parameters.

    do(deed, **iops)

    Inherited Class Attributes:
        Registry (dict): subclass registry whose items are (name, cls) where:
                name is unique name for subclass
                cls is reference to class object
        Instances (dict): instance registry whose items are (name, instance) where:
                name is unique instance name and instance is instance reference
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.
        Names (tuple[str]): tuple of aliases (names) under which this subclas
                            appears in .Registry. Created by @register

    Overridden Class Attributes
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.


    Inherited Properties:
        name (str): unique name string of instance
        iops (dict): input-output-parameters for .act
        context (str): action context for .act

    Inherited Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork

    Attributes:
        deed (Callable): action to be called with .iops as parameters else
                executable set of statements with M and D as locals

    Properties:
        compiled (bool): True means ._code holds compiled .deed
                         False means not yet compiled


    Hidden:

        _code (None|CodeType): compiled evalable boolean expression .expr
            None means not yet compiled from .expr


    Hidden
        _name (str|None): unique name of instance
        _iopts (dict): input-output-paramters for .act
        _context (str): action context for .act
        _code (CodeType): compiled executable set of statements that execs .deed
            when it is a noncallable str.  M and D are in the locals of the exec.

    """
    Index = 0  # naming index for default names of this subclasses instances
    #Names = () tuple of aliases for this subclass created by @register



    def __init__(self, deed=None, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            name (str|None): unique name of this instance. When None then
                generate name from .Index
            iops (dict|None): input-output-parameters for .act. When None then
                set to empty dict.
            context (str|None): action context for .act. Default is "enter"
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork

        Parameters:
            deed (None|Callable):  callable to be actioned with iops

        """
        super(Act, self).__init__(**kwa)
        self.deed = deed if deed is not None else (lambda **iops: iops)
        self.iops.update(M=self.mine, D=self.dock)  # inject .mine and .dock
        self._code = None


    def act(self, **iops):  # passed in by call
        """Act called by ActBase."""
        if callable(self.deed):
            return self.deed(**iops)

        if not self.compiled:  # not yet compiled so lazy
            self.compile()  # first time only recompile to ._code
        M = self.mine  # ensure M is in locals() for exec
        D = self.dock  # ensure D is in locals() for exec
        # note iops already in locals() for exec
        return exec(self._code)



    @property
    def compiled(self):
        """Property compiled

        Returns:
            compiled (bool): True means ._code holds compiled ._expr
                             False means not yet compiled
        """
        return True if self._code is not None else False


    def compile(self):
        """Compile evable boolean expression str ._expr into compiled code
        object ._code to be evaluated at run time.
        Because code objects are not pickleable the compilation must happen
        at prep (enter) time not init time.
        """
        self._code = compile(self.deed, '<string>', 'exec')




@register()
class Tract(ActBase):
    """Tract (transit act) is subclass of ActBase whose .act evaluates conditional
    need expression to determine if a transition condition is satified for
    transition to its destination box.

    Inherited Class Attributes:
        Registry (dict): subclass registry whose items are (name, cls) where:
                name is unique name for subclass
                cls is reference to class object
        Instances (dict): instance registry whose items are (name, instance) where:
                name is unique instance name and instance is instance reference
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.
        Names (tuple[str]): tuple of aliases (names) under which this subclas
                            appears in .Registry. Created by @register

    Overridden Class Attributes
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.

    Inherited Properties:
        name (str): unique name string of instance
        iops (dict): input-output-parameters for .act
        context (str): action context for .act

    Inherited Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork

    Attributes:
        dest (Box): destination Box for this transition.
        need (Need): transition condition to be evaluated

    Hidden
        _name (str|None): unique name of instance
        _iopts (dict): input-output-paramters for .act
        _context (str): action context for .act

    """
    Index = 0  # naming index for default names of this subclasses instances
    #Names = () tuple of aliases for this subclass created by @register



    def __init__(self, dest=None, need=None, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            name (str|None): unique name of this instance. When None then
                generate name from .Index
            iops (dict|None): input-output-parameters for .act. When None then
                set to empty dict.
            context (str): action context for .act
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork

        Parameters:
            dest (None|str|Box): destination Box for this transition.
                When None then resolve later to next box of current box
                When str is box name then resolve to box with that name
                When Box instance then use directly
            need (None|str|Need): transition condition to be evaluated
                When None then always evaluates to True
                When str = bool expression then create Need from expression
                When Need instance then use directly

        """
        kwa.update(context=Context.transit)  # override must be transit context
        super(Tract, self).__init__(**kwa)
        self.dest = dest if dest is not None else 'next'  # default is next
        self.need = need if need is not None else Need()  # default need evals to True
        if self.context != Context.transit:
            raise HierError(f"Invalid context='{self.context}' for Tract "
                            f"'{self.name}'")



    def act(self, **iops):
        """Act called by ActBase."""
        if self.need():
            if not isinstance(self.dest, boxing.Box):
                raise HierError(f"Unresolved dest={self.dest}")
            return self.dest
        else:
            return None


@register(names=('end', 'End'))
class EndAct(ActBase):
    """EndAct is subclass of ActBase whose .act indicates a desire to end the
    boxer by setting bag at .iops "end" .value to True. Where "end" is at key
    "_boxer_boxername_end".



    Inherited Class Attributes:
        Registry (dict): subclass registry whose items are (name, cls) where:
                name is unique name for subclass
                cls is reference to class object
        Instances (dict): instance registry whose items are (name, instance) where:
                name is unique instance name and instance is instance reference
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.
        Names (tuple[str]): tuple of aliases (names) under which this subclas
                            appears in .Registry. Created by @register

    Overridden Class Attributes
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.


    Inherited Properties:
        name (str): unique name string of instance
        iops (dict): input-output-parameters for .act
        context (str): action context for .act

    Inherited Attributes:
        mine (Mine): ephemeral bags in mine (in memory) shared by boxwork
        dock (Dock): durable bags in dock (on disc) shared by boxwork

    Attributes:

    Used iops:
        _boxer (str):  boxer name


    Hidden
        _name (str|None): unique name of instance
        _iopts (dict): input-output-paramters for .act
        _context (str): action context for .act

    """
    Index = 0  # naming index for default names of this subclasses instances
    #Names = () tuple of aliases for this subclass created by @register



    def __init__(self, context=Context.enter, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            name (str|None): unique name of this instance. When None then
                generate name from .Index
            iops (dict|None): input-output-parameters for .act. When None then
                set to empty dict.
            context (str): action context for .act
            mine (None|Mine): ephemeral bags in mine (in memory) shared by boxwork
            dock (None|Dock): durable bags in dock (on disc) shared by boxwork

        Parameters:

        Used iops:
            _boxer (str): boxer name. Implicit iop injected by verb (do etc)


        """
        super(EndAct, self).__init__(context=context, **kwa)

        try:
            boxer = self.iops['_boxer']  # get boxer name
        except KeyError as ex:
            raise HierError(f"Missing iops for '{self.name}' instance "
                            f"of Act self.__class__.__name__") from ex


        keys = ("", "boxer", boxer, "end")  # _boxer_boxername_end
        if keys not in self.mine:
            self.mine[keys] = Bag()  # create bag at end default value = None


    def act(self, **iops):
        """Act called by ActBase."""
        boxer = self.iops['_boxer']  # get boxer name
        keys = ("", "boxer", boxer, "end")
        self.mine[keys].value = True
